unit KOLCxGif;
{ This GIF implementation is converted to Pascal manually by Vladimir Kladov, 2007
  from CxImage (ximagif.h, ximagif.cpp). See original library at http://www.xdp.com
}

interface

uses Windows, KOL;

const
//* Various error codes used by decoder */
  OUT_OF_MEMORY   = -10;
  BAD_CODE_SIZE   = -20;
  READ_ERROR      = -1;
  WRITE_ERROR     = -2;
  OPEN_ERROR      = -3;
  CREATE_ERROR    = -4;
  MAX_CODES       = 4095;
  GIFBUFTAM       = 16383;
  TRANSPARENCY_CODE = $F9;

//LZW GIF Image compression
  MAXBITSCODES    = 12;
  HSIZE           = 5003;     //* 80% occupancy */
  //--HashTabOf(i)    = htab[i];
  //--CodeTabOf(i)    = codetab[i];

type
  Tgifgce = packed record
    flags: Byte;
    delaytime: Word;
    transpcolorindex: Byte;
  end;

  Tdscgif = packed record           // Logic screen descriptor
    header: array[ 0..5 ] of Char;  // Firma and version
    scrwidth, scrheight: Word;
    pflds, bcindx, pxasrat: Char;
  end;

  Ttag_image = packed record
    l, t, w, h: Word;
  end;

  Trgb = packed record
    r, g, b: Byte;
  end;

  TTabCol = packed record           // Tabla de colores
    colres,                         // color resolution
    sogct: SmallInt;                // size of global color table
    paleta: array[ 0..255 ] of TRgb;// paleta
  end;

  Ttag_RLE = packed record
    rl_pixel,
    rl_basecode,
    rl_count,
    rl_table_pixel,
    rl_table_max,
    just_cleared,
    out_bits,
    out_bits_init,
    out_count,
    out_bump,
    out_bump_init,
    out_clear,
    out_clear_init,
    max_ocodes,
    code_clear,
    code_eof: Integer;
    obuf: DWORD;
    obits: Integer;
    oblock: array[ 0..255 ] of Byte;
    oblen: Integer;
  end;

  PGifImage = ^TGifImage;

  PImageIterator = ^TImageIterator;
  TImageIterator = object( TObj )
  private
    procedure SetY(const Value: Integer);
  public
    Itx, Ity: Integer;
    Stepx, Stepy: Integer;
    IterImage: PByte;
    ima: PGifImage;
  public
    function ItOK: Boolean;
    procedure Reset;
    procedure Upset;
    function NextRow: Boolean;
    function PrevRow: Boolean;
    property Y: Integer read Ity write SetY;
    procedure SetRow( buf: PByte; n: Integer );
    procedure GetRow( buf: PByte; n: Integer );
    function NextByte: Boolean;
    function PrevByte: Boolean;
    function NextStep: Boolean;
    function PrevStep: Boolean;
    function GetCol( pCol: PByte; x: Integer ): Boolean;
    function SetCol( pCol: pByte; x: Integer ): Boolean;
  end;

  TGifImage = object( TObj )
  protected
    fBitmap: PBitmap;
    function IsInside( x, y: Integer ): Boolean;
    function BPP: Integer;
  public

  end;

  PKOLGif = ^TKOLGif;
  TKOLGif = object( TObj )
  end;

implementation

function MAXCODE(n_bits: Integer): Integer;
begin
  Result := (1 shl n_bits) - 1;
end;

{ TImageIterator }

{ TGifImage }

function TGifImage.BPP: Integer;
begin
  CASE fBitmap.PixelFormat OF
  pf1bit: Result := 1;
  pf4bit: Result := 4;
  pf8bit: Result := 8;
  pf15bit, pf16bit: Result := 16;
  pf24bit: Result := 24;
  pf32bit: Result := 32;
  else Result := 0;
  END;
end;

function TGifImage.IsInside(x, y: Integer): Boolean;
begin
  Result := (fBitmap <> nil) and
            (x >= 0) and (x < fBitmap.Width) and
            (y >= 0) and (y < fBitmap.Height);
end;

{ TImageIterator }

function TImageIterator.GetCol(pCol: PByte; x: Integer): Boolean;
var i, yy, bytes: Integer;
    pSrc: PByte;
begin
  Result := FALSE;
  if (pCol = nil) or (ima.BPP < 8) or (x > ima.fBitmap.Width) then
    Exit;
  bytes := ima.BPP shr 3;
  for yy := 0 to ima.fBitmap.Height-1 do
  begin
    pSrc := Pointer( Integer( ima.fBitmap.ScanLine[ yy ] ) + x * bytes );
    for i := 1 to bytes do
    begin
      pCol^ := pSrc^;
      inc( pCol );
      inc( pSrc );
    end;
  end;
  Result := TRUE;
end;

procedure TImageIterator.GetRow(buf: PByte; n: Integer);
begin
  if (IterImage <> nil) and (buf <> nil) and (n > 0) then
    move( IterImage^, buf^, n );
end;

function TImageIterator.ItOK: Boolean;
begin
  if ima <> nil then Result := ima.IsInside( Itx, Ity )
  else Result := FALSE;
end;

function TImageIterator.NextByte: Boolean;
begin
  Result := TRUE;
  Inc( Itx );
  if Itx < ima.fBitmap.Width then Exit;
  inc( Ity );
  if Ity <= ima.fBitmap.Height then
  begin
    Itx := 0;
    IterImage := ima.fBitmap.ScanLine[ Ity ];
  end
    else Result := FALSE;
end;

function TImageIterator.NextRow: Boolean;
begin
  Result := FALSE;
  inc( Ity );
  if Ity >= ima.fBitmap.Height then Exit;
  IterImage := ima.fBitmap.ScanLine[ Ity ];
  Result := TRUE;
end;

function TImageIterator.NextStep: Boolean;
begin
  Result := TRUE;
  inc( Itx, Stepx );
  if Itx < ima.fBitmap.Width then Exit;
  inc( Ity, Stepy );
  if Ity < ima.fBitmap.Height then
  begin
    IterImage := ima.fBitmap.ScanLine[ Ity ];
    Itx := 0;
  end
    else Result := FALSE;
end;

function TImageIterator.PrevByte: Boolean;
begin
  Result := TRUE;
  dec( Itx );
  if Itx >= 0 then Exit;
  dec( Ity );
  if Ity >= 0 then
  begin
    IterImage := ima.fBitmap.ScanLine[ Ity ];
    Itx := 0;
  end
    else Result := FALSE;
end;

function TImageIterator.PrevRow: Boolean;
begin
  Result := FALSE;
  dec( Ity );
  if Ity < 0 then Exit;
  IterImage := ima.fBitmap.ScanLine[ Ity ];
  Result := TRUE;
end;

function TImageIterator.PrevStep: Boolean;
begin
  Result := TRUE;
  dec( Itx, Stepx );
  if Itx >= 0 then Exit;
  dec( Ity, Stepy );
  if Ity >= 0 then
  begin
    IterImage := ima.fBitmap.ScanLine[ Ity ];
    Itx := 0;
  end
    else Result := FALSE;
end;

procedure TImageIterator.Reset;
begin
  if ima <> nil then IterImage := ima.FBitmap.ScanLine[ 0 ]
  else               IterImage := nil;
  Itx := 0;
  Ity := 0;
end;

function TImageIterator.SetCol(pCol: pByte; x: Integer): Boolean;
var i, yy, bytes: Integer;
    pDst: PByte;
begin
  Result := FALSE;
  if (pCol = nil) or (ima.BPP < 8) or (x > ima.fBitmap.Width) then
    Exit;
  bytes := ima.BPP shr 3;
  for yy := 0 to ima.fBitmap.Height-1 do
  begin
    pDst := Pointer( Integer( ima.fBitmap.ScanLine[ yy ] ) + x * bytes );
    for i := 1 to bytes do
    begin
      pDst^ := pCol^;
      inc( pCol );
      inc( pDst );
    end;
  end;
  Result := TRUE;
end;

procedure TImageIterator.SetRow(buf: PByte; n: Integer);
begin
  if n < 0 then n := ima.fBitmap.ScanLineSize
  else n := min( n, ima.fBitmap.ScanLineSize );
  if (IterImage <> nil) and (buf <> nil) and (n > 0) then
    move( buf^, IterImage^, n );
end;

procedure TImageIterator.SetY(const Value: Integer);
begin
  if (Value < 0) or (Value >= ima.fBitmap.Height) then Exit;
  Ity := Value;
  IterImage := ima.fBitmap.ScanLine[ Ity ];
end;

procedure TImageIterator.Upset;
begin
  Itx := 0;
  Ity := ima.fBitmap.Height-1;
  IterImage := ima.fBitmap.ScanLine[ Ity ];
end;

end.
