unit KOLTIFF;

{ Converted by Vladimir Kladov, 2007 from CxImage graphic encoding and decoding
  library. Though decode only implemented for tiff, since there were necessary
  to encode tiff images for me.

  This unit based on CxImage library (unit ximatif.cpp). All other units
  except named ex_***.pas are obtained using my Casm2Basm conversion utility
  from the assempler output got from MS VC++ (with small changes: for 3 of
  declarations, the third parameter p3: Pointer was added manually at the place
  of variable amount of parameters list).

  This object allows opening fast very big tiff images, and it understands a lot
  of tiff format variations, staying very stable and fast (in comparison with
  GDI+, KOLGraphic and LibTiffDelphi). Thanks a lot to the CxImage developer
  Davide Pizzolato! (See also License.txt attached,
  which is originated from CxImage).
}

interface

uses Windows, KOL, kolmath, KOLJpeg;

{$define FIX_16BPP_DARKIMG}

//{$define test_SetField}

const
  bigendian = FALSE;

  IMAGE_MAX_MEMORY = 256000000;

  TIFF_FILLORDER    = $0003;	//* natural bit fill order for machine */
  TIFF_DIRTYHEADER  = $0004;	//* header must be written on close */
  TIFF_DIRTYDIRECT  = $0008;	//* current directory must be written */
  TIFF_BUFFERSETUP  = $0010;	//* data buffers setup */
  TIFF_CODERSETUP   = $0020;	//* encoder/decoder setup done */
  TIFF_BEENWRITING  = $0040;	//* written 1+ scanlines to file */
  TIFF_SWAB	    = $0080;	//* byte swap file information */
  TIFF_NOBITREV	    = $0100;	//* inhibit bit reversal logic */
  TIFF_MYBUFFER	    = $0200;	//* my raw data buffer; free on close */
  TIFF_ISTILED	    = $0400;	//* file is tile, not strip- based */
  TIFF_MAPPED	    = $0800;	//* file is mapped into memory */
  TIFF_POSTENCODE   = $1000;	//* need call to postencode routine */
  TIFF_INSUBIFD	    = $2000;	//* currently writing a subifd */
  TIFF_UPSAMPLED    = $4000;	//* library is doing data up-sampling */
  TIFF_STRIPCHOP    = $8000;	//* enable strip chopping support */

  FIELD_SETLONGS    =	3;

  TIFF_VERSION	    = 42;
  TIFF_BIGENDIAN    = $4d4d;
  TIFF_LITTLEENDIAN = $4949;


(*
 * TIFF Tag Definitions.
 *)
  TIFFTAG_SUBFILETYPE =		254;	//* subfile data descriptor */
  FILETYPE_REDUCEDIMAGE	= $1	;       //* reduced resolution version */
  FILETYPE_PAGE		= $2	;       //* one page of many */
  FILETYPE_MASK		= $4	;       //* transparency mask */
  TIFFTAG_OSUBFILETYPE		= 255	; //* +kind of data in subfile */
  OFILETYPE_IMAGE		= 1	; //* full resolution image data */
  OFILETYPE_REDUCEDIMAGE	= 2	; //* reduced size image data */
  OFILETYPE_PAGE		= 3	; //* one page of many */
  TIFFTAG_IMAGEWIDTH		= 256	; //* image width in pixels */
  TIFFTAG_IMAGELENGTH		= 257	; //* image height in pixels */
  TIFFTAG_BITSPERSAMPLE		= 258	; //* bits per channel (sample) */
  TIFFTAG_COMPRESSION		= 259	; //* data compression technique */
  COMPRESSION_NONE		= 1	; //* dump mode */
  COMPRESSION_CCITTRLE	= 2	;         //* CCITT modified Huffman RLE */
  COMPRESSION_CCITTFAX3	= 3	;         //* CCITT Group 3 fax encoding */
  COMPRESSION_CCITT_T4        = 3       ; //* CCITT T.4 (TIFF 6 name) */
  COMPRESSION_CCITTFAX4	= 4	;         //* CCITT Group 4 fax encoding */
  COMPRESSION_CCITT_T6        = 4       ; //* CCITT T.6 (TIFF 6 name) */
  COMPRESSION_LZW		= 5       ; //* Lempel-Ziv  & Welch */
  COMPRESSION_OJPEG		= 6	; //* !6.0 JPEG */
  COMPRESSION_JPEG		= 7	; //* %JPEG DCT compression */
  COMPRESSION_NEXT		= 32766	; //* NeXT 2-bit RLE */
  COMPRESSION_CCITTRLEW	= 32771	;         //* #1 w/ word alignment */
  COMPRESSION_PACKBITS	= 32773	;         //* Macintosh RLE */
  COMPRESSION_THUNDERSCAN	= 32809	; //* ThunderScan RLE */
//* codes 32895-32898 are reserved for ANSI IT8 TIFF/IT <dkelly@etsinc.com) */
  COMPRESSION_IT8CTPAD	= 32895   ; //* IT8 CT w/padding */
  COMPRESSION_IT8LW		= 32896   ; //* IT8 Linework RLE */
  COMPRESSION_IT8MP		= 32897   ; //* IT8 Monochrome picture */
  COMPRESSION_IT8BL		= 32898   ; //* IT8 Binary line art */
//* compression codes 32908-32911 are reserved for Pixar */
  COMPRESSION_PIXARFILM	= 32908   ;       //* Pixar companded 10bit LZW */
  COMPRESSION_PIXARLOG	= 32909   ;       //* Pixar companded 11bit ZIP */
  COMPRESSION_DEFLATE		= 32946	; //* Deflate compression */
  COMPRESSION_ADOBE_DEFLATE   = 8       ; //* Deflate compression, as recognized by Adobe */
//* compression code 32947 is reserved for Oceana Matrix <dev@oceana.com> */
  COMPRESSION_DCS             = 32947   ; //* Kodak DCS encoding */
  COMPRESSION_JBIG		= 34661	; //* ISO JBIG */
  COMPRESSION_SGILOG		= 34676	; //* SGI Log Luminance RLE */
  COMPRESSION_SGILOG24	= 34677	;         //* SGI Log 24-bit packed */
  TIFFTAG_PHOTOMETRIC		= 262	; //* photometric interpretation */
  PHOTOMETRIC_MINISWHITE	= 0	; //* min value is white */
  PHOTOMETRIC_MINISBLACK	= 1	; //* min value is black */
  PHOTOMETRIC_RGB		= 2	; //* RGB color model */
  PHOTOMETRIC_PALETTE		= 3	; //* color map indexed */
  PHOTOMETRIC_MASK		= 4	; //* $holdout mask */
  PHOTOMETRIC_SEPARATED	= 5	;         //* !color separations */
  PHOTOMETRIC_YCBCR		= 6	; //* !CCIR 601 */
  PHOTOMETRIC_CIELAB		= 8	; //* !1976 CIE L*a*b* */
  PHOTOMETRIC_ITULAB		= 10	; //* ITU L*a*b* */
  PHOTOMETRIC_LOGL		= 32844	; //* CIE Log2(L) */
  PHOTOMETRIC_LOGLUV		= 32845	; //* CIE Log2(L) (u',v') */
  TIFFTAG_THRESHHOLDING		= 263	; //* +thresholding used on data */
  THRESHHOLD_BILEVEL		= 1	; //* b&w art scan */
  THRESHHOLD_HALFTONE		= 2	; //* or dithered scan */
  THRESHHOLD_ERRORDIFFUSE	= 3	; //* usually floyd-steinberg */
  TIFFTAG_CELLWIDTH		= 264	; //* +dithering matrix width */
  TIFFTAG_CELLLENGTH		= 265	; //* +dithering matrix height */
  TIFFTAG_FILLORDER		= 266	; //* data order within a byte */
  FILLORDER_MSB2LSB		= 1	; //* most significant -> least */
  FILLORDER_LSB2MSB		= 2	; //* least significant -> most */
  TIFFTAG_DOCUMENTNAME		= 269	; //* name of doc. image is from */
  TIFFTAG_IMAGEDESCRIPTION	= 270	; //* info about image */
  TIFFTAG_MAKE			= 271	; //* scanner manufacturer name */
  TIFFTAG_MODEL			= 272	; //* scanner model name/number */
  TIFFTAG_STRIPOFFSETS		= 273	; //* offsets to data strips */
  TIFFTAG_ORIENTATION		= 274	; //* +image orientation */
  ORIENTATION_TOPLEFT		= 1	; //* row 0 top, col 0 lhs */
  ORIENTATION_TOPRIGHT	= 2	;         //* row 0 top, col 0 rhs */
  ORIENTATION_BOTRIGHT	= 3	;         //* row 0 bottom, col 0 rhs */
  ORIENTATION_BOTLEFT		= 4	; //* row 0 bottom, col 0 lhs */
  ORIENTATION_LEFTTOP		= 5	; //* row 0 lhs, col 0 top */
  ORIENTATION_RIGHTTOP	= 6	;         //* row 0 rhs, col 0 top */
  ORIENTATION_RIGHTBOT	= 7	;         //* row 0 rhs, col 0 bottom */
  ORIENTATION_LEFTBOT		= 8	; //* row 0 lhs, col 0 bottom */
  TIFFTAG_SAMPLESPERPIXEL		= 277	; //* samples per pixel */
  TIFFTAG_ROWSPERSTRIP		= 278	; //* rows per strip of data */
  TIFFTAG_STRIPBYTECOUNTS		= 279	; //* bytes counts for strips */
  TIFFTAG_MINSAMPLEVALUE		= 280	; //* +minimum sample value */
  TIFFTAG_MAXSAMPLEVALUE		= 281	; //* +maximum sample value */
  TIFFTAG_XRESOLUTION		= 282	; //* pixels/resolution in x */
  TIFFTAG_YRESOLUTION		= 283	; //* pixels/resolution in y */
  TIFFTAG_PLANARCONFIG		= 284	; //* storage organization */
  PLANARCONFIG_CONTIG		= 1	; //* single image plane */
  PLANARCONFIG_SEPARATE	= 2	;         //* separate planes of data */
  TIFFTAG_PAGENAME		= 285	; //* page name image is from */
  TIFFTAG_XPOSITION		= 286	; //* x page offset of image lhs */
  TIFFTAG_YPOSITION		= 287	; //* y page offset of image lhs */
  TIFFTAG_FREEOFFSETS		= 288	; //* +byte offset to free block */
  TIFFTAG_FREEBYTECOUNTS		= 289	; //* +sizes of free blocks */
  TIFFTAG_GRAYRESPONSEUNIT	= 290	; //* $gray scale curve accuracy */
  GRAYRESPONSEUNIT_10S	= 1	; //* tenths of a unit */
  GRAYRESPONSEUNIT_100S	= 2	; //* hundredths of a unit */
  GRAYRESPONSEUNIT_1000S	= 3	; //* thousandths of a unit */
  GRAYRESPONSEUNIT_10000S	= 4	; //* ten-thousandths of a unit */
  GRAYRESPONSEUNIT_100000S	= 5	; //* hundred-thousandths */
  TIFFTAG_GRAYRESPONSECURVE	= 291	; //* $gray scale response curve */
  TIFFTAG_GROUP3OPTIONS		= 292	; //* 32 flag bits */
  TIFFTAG_T4OPTIONS		= 292	; //* TIFF 6.0 proper name alias */
  GROUP3OPT_2DENCODING	= $1	; //* 2-dimensional coding */
  GROUP3OPT_UNCOMPRESSED	= $2	; //* data not compressed */
  GROUP3OPT_FILLBITS		= $4	; //* fill to byte boundary */
  TIFFTAG_GROUP4OPTIONS		= 293	; //* 32 flag bits */
  TIFFTAG_T6OPTIONS               = 293     ; //* TIFF 6.0 proper name */
  GROUP4OPT_UNCOMPRESSED	= $2	; //* data not compressed */
  TIFFTAG_RESOLUTIONUNIT		= 296	; //* units of resolutions */
  RESUNIT_NONE		= 1	; //* no meaningful units */
  RESUNIT_INCH		= 2	; //* english */
  RESUNIT_CENTIMETER		= 3	; //* metric */
  TIFFTAG_PAGENUMBER		= 297	; //* page numbers of multi-page */
  TIFFTAG_COLORRESPONSEUNIT	= 300	; //* $color curve accuracy */
  COLORRESPONSEUNIT_10S	= 1	; //* tenths of a unit */
  COLORRESPONSEUNIT_100S	= 2	; //* hundredths of a unit */
  COLORRESPONSEUNIT_1000S	= 3	; //* thousandths of a unit */
  COLORRESPONSEUNIT_10000S	= 4	; //* ten-thousandths of a unit */
  COLORRESPONSEUNIT_100000S	= 5	; //* hundred-thousandths */
  TIFFTAG_TRANSFERFUNCTION	= 301	; //* !colorimetry info */
  TIFFTAG_SOFTWARE		= 305	; //* name & release */
  TIFFTAG_DATETIME		= 306	; //* creation date and time */
  TIFFTAG_ARTIST			= 315	; //* creator of image */
  TIFFTAG_HOSTCOMPUTER		= 316	; //* machine where created */
  TIFFTAG_PREDICTOR		= 317	; //* prediction scheme w/ LZW */
  TIFFTAG_WHITEPOINT		= 318	; //* image white point */
  TIFFTAG_PRIMARYCHROMATICITIES	= 319	; //* !primary chromaticities */
  TIFFTAG_COLORMAP		= 320	; //* RGB map for pallette image */
  TIFFTAG_HALFTONEHINTS		= 321	; //* !highlight+shadow info */
  TIFFTAG_TILEWIDTH		= 322	; //* !rows/data tile */
  TIFFTAG_TILELENGTH		= 323	; //* !cols/data tile */
  TIFFTAG_TILEOFFSETS		= 324	; //* !offsets to data tiles */
  TIFFTAG_TILEBYTECOUNTS		= 325	; //* !byte counts for tiles */
  TIFFTAG_BADFAXLINES		= 326	; //* lines w/ wrong pixel count */
  TIFFTAG_CLEANFAXDATA		= 327	; //* regenerated line info */
  CLEANFAXDATA_CLEAN		= 0	; //* no errors detected */
  CLEANFAXDATA_REGENERATED	= 1	; //* receiver regenerated lines */
  CLEANFAXDATA_UNCLEAN	= 2	; //* uncorrected errors exist */
  TIFFTAG_CONSECUTIVEBADFAXLINES	= 328	; //* max consecutive bad lines */
  TIFFTAG_SUBIFD			= 330	; //* subimage descriptors */
  TIFFTAG_INKSET			= 332	; //* !inks in separated image */
  INKSET_CMYK			= 1	; //* !cyan-magenta-yellow-black */
  TIFFTAG_INKNAMES		= 333	; //* !ascii names of inks */
  TIFFTAG_NUMBEROFINKS		= 334	; //* !number of inks */
  TIFFTAG_DOTRANGE		= 336	; //* !0% and 100% dot codes */
  TIFFTAG_TARGETPRINTER		= 337	; //* !separation target */
  TIFFTAG_EXTRASAMPLES		= 338	; //* !info about extra samples */
  EXTRASAMPLE_UNSPECIFIED	= 0	; //* !unspecified data */
  EXTRASAMPLE_ASSOCALPHA	= 1	; //* !associated alpha data */
  EXTRASAMPLE_UNASSALPHA	= 2	; //* !unassociated alpha data */
  TIFFTAG_SAMPLEFORMAT		= 339	; //* !data sample format */
  SAMPLEFORMAT_UINT		= 1	; //* !unsigned integer data */
  SAMPLEFORMAT_INT		= 2	; //* !signed integer data */
  SAMPLEFORMAT_IEEEFP		= 3	; //* !IEEE floating point data */
  SAMPLEFORMAT_VOID		= 4	; //* !untyped data */
  SAMPLEFORMAT_COMPLEXINT	= 5	; //* !complex signed int */
  SAMPLEFORMAT_COMPLEXIEEEFP	= 6	; //* !complex ieee floating */
  TIFFTAG_SMINSAMPLEVALUE		= 340	; //* !variable MinSampleValue */
  TIFFTAG_SMAXSAMPLEVALUE		= 341	; //* !variable MaxSampleValue */
  TIFFTAG_JPEGTABLES		= 347	; //* %JPEG table stream */
(*
 * Tags 512-521 are obsoleted by Technical Note #2
 * which specifies a revised JPEG-in-TIFF scheme.
 *)
  TIFFTAG_JPEGPROC		= 512	; //* !JPEG processing algorithm */
  JPEGPROC_BASELINE		= 1	; //* !baseline sequential */
  JPEGPROC_LOSSLESS		= 14	; //* !Huffman coded lossless */
  TIFFTAG_JPEGIFOFFSET		= 513	; //* !pointer to SOI marker */
  TIFFTAG_JPEGIFBYTECOUNT		= 514	; //* !JFIF stream length */
  TIFFTAG_JPEGRESTARTINTERVAL	= 515	; //* !restart interval length */
  TIFFTAG_JPEGLOSSLESSPREDICTORS	= 517	; //* !lossless proc predictor */
  TIFFTAG_JPEGPOINTTRANSFORM	= 518	; //* !lossless point transform */
  TIFFTAG_JPEGQTABLES		= 519	; //* !Q matrice offsets */
  TIFFTAG_JPEGDCTABLES		= 520	; //* !DCT table offsets */
  TIFFTAG_JPEGACTABLES		= 521	; //* !AC coefficient offsets */
  TIFFTAG_YCBCRCOEFFICIENTS	= 529	; //* !RGB -> YCbCr transform */
  TIFFTAG_YCBCRSUBSAMPLING	= 530	; //* !YCbCr subsampling factors */
  TIFFTAG_YCBCRPOSITIONING	= 531	; //* !subsample positioning */
  YCBCRPOSITION_CENTERED	= 1	; //* !as in PostScript Level 2 */
  YCBCRPOSITION_COSITED	= 2	; //* !as in CCIR 601-1 */
  TIFFTAG_REFERENCEBLACKWHITE	= 532	; //* !colorimetry info */
//* tags 32952-32956 are private tags registered to Island Graphics */
  TIFFTAG_REFPTS			= 32953	; //* image reference points */
  TIFFTAG_REGIONTACKPOINT		= 32954	; //* region-xform tack point */
  TIFFTAG_REGIONWARPCORNERS	= 32955	; //* warp quadrilateral */
  TIFFTAG_REGIONAFFINE		= 32956	; //* affine transformation mat */
//* tags 32995-32999 are private tags registered to SGI */
  TIFFTAG_MATTEING		= 32995	; //* $use ExtraSamples */
  TIFFTAG_DATATYPE		= 32996	; //* $use SampleFormat */
  TIFFTAG_IMAGEDEPTH		= 32997	; //* z depth of image */
  TIFFTAG_TILEDEPTH		= 32998	; //* z depth/data tile */
//* tags 33300-33309 are private tags registered to Pixar */
(*
 * TIFFTAG_PIXAR_IMAGEFULLWIDTH and TIFFTAG_PIXAR_IMAGEFULLLENGTH
 * are set when an image has been cropped out of a larger image.
 * They reflect the size of the original uncropped image.
 * The TIFFTAG_XPOSITION and TIFFTAG_YPOSITION can be used
 * to determine the position of the smaller image in the larger one.
 *)
  TIFFTAG_PIXAR_IMAGEFULLWIDTH    = 33300   ; //* full image size in x */
  TIFFTAG_PIXAR_IMAGEFULLLENGTH   = 33301   ; //* full image size in y */
 (* Tags 33302-33306 are used to identify special image modes and data
  * used by Pixar's texture formats.
  *)
  TIFFTAG_PIXAR_TEXTUREFORMAT	= 33302	; //* texture map format */
  TIFFTAG_PIXAR_WRAPMODES		= 33303	; //* s & t wrap modes */
  TIFFTAG_PIXAR_FOVCOT		= 33304	; //* cotan(fov) for env. maps */
  TIFFTAG_PIXAR_MATRIX_WORLDTOSCREEN = 33305;
  TIFFTAG_PIXAR_MATRIX_WORLDTOCAMERA = 33306;
//* tag 33405 is a private tag registered to Eastman Kodak */
  TIFFTAG_WRITERSERIALNUMBER      = 33405   ; //* device serial number */
//* tag 33432 is listed in the 6.0 spec w/ unknown ownership */
  TIFFTAG_COPYRIGHT		= 33432	; //* copyright string */
//* IPTC TAG from RichTIFF specifications */
  TIFFTAG_RICHTIFFIPTC    = 33723;
//* 34016-34029 are reserved for ANSI IT8 TIFF/IT <dkelly@etsinc.com) */
  TIFFTAG_IT8SITE			= 34016	; //* site name */
  TIFFTAG_IT8COLORSEQUENCE	= 34017	; //* color seq. [RGB,CMYK,etc] */
  TIFFTAG_IT8HEADER		= 34018	; //* DDES Header */
  TIFFTAG_IT8RASTERPADDING	= 34019	; //* raster scanline padding */
  TIFFTAG_IT8BITSPERRUNLENGTH	= 34020	; //* # of bits in short run */
  TIFFTAG_IT8BITSPEREXTENDEDRUNLENGTH = 34021; //* # of bits in long run */
  TIFFTAG_IT8COLORTABLE		= 34022	; //* LW colortable */
  TIFFTAG_IT8IMAGECOLORINDICATOR	= 34023	; //* BP/BL image color switch */
  TIFFTAG_IT8BKGCOLORINDICATOR	= 34024	; //* BP/BL bg color switch */
  TIFFTAG_IT8IMAGECOLORVALUE	= 34025	; //* BP/BL image color value */
  TIFFTAG_IT8BKGCOLORVALUE	= 34026	; //* BP/BL bg color value */
  TIFFTAG_IT8PIXELINTENSITYRANGE	= 34027	; //* MP pixel intensity value */
  TIFFTAG_IT8TRANSPARENCYINDICATOR = 34028	; //* HC transparency switch */
  TIFFTAG_IT8COLORCHARACTERIZATION = 34029	; //* color character. table */
//* tags 34232-34236 are private tags registered to Texas Instruments */
  TIFFTAG_FRAMECOUNT              = 34232   ; //* Sequence Frame Count */
//* tag 34750 is a private tag registered to Adobe? */
  TIFFTAG_ICCPROFILE		= 34675	; //* ICC profile data */
//* tag 34377 is private tag registered to Adobe for PhotoShop */
  TIFFTAG_PHOTOSHOP				= 34377;
//* tag 34750 is a private tag registered to Pixel Magic */
  TIFFTAG_JBIGOPTIONS		= 34750	; //* JBIG options */
//* tags 34908-34914 are private tags registered to SGI */
  TIFFTAG_FAXRECVPARAMS		= 34908	; //* encoded Class 2 ses. parms */
  TIFFTAG_FAXSUBADDRESS		= 34909	; //* received SubAddr string */
  TIFFTAG_FAXRECVTIME		= 34910	; //* receive time (secs) */
//* tags 37439-37443 are registered to SGI <gregl@sgi.com> */
  TIFFTAG_STONITS			= 37439	; //* Sample value to Nits */
//* tag 34929 is a private tag registered to FedEx */
  TIFFTAG_FEDEX_EDR		= 34929	; //* unknown use */
//* tag 65535 is an undefined tag used by Eastman Kodak */
  TIFFTAG_DCSHUESHIFTVALUES       = 65535   ; //* hue shift correction data */
(*
 * The following are ``pseudo tags'' that can be
 * used to control codec-specific functionality.
 * These tags are not written to file.  Note that
 * these values start at $ffff+1 so that they'll
 * never collide with Aldus-assigned tags.
 *
 * If you want your private pseudo tags ``registered''
 * (i.e. added to this file), send mail to sam@sgi.com
 * with the appropriate C definitions to add.
 *)
  TIFFTAG_FAXMODE			= 65536	; //* Group 3/4 format control */
  FAXMODE_CLASSIC	= $0000		; //* default, include RTC */
  FAXMODE_NORTC	= $0001		; //* no RTC at end of data */
  FAXMODE_NOEOL	= $0002		; //* no EOL code at end of row */
  FAXMODE_BYTEALIGN	= $0004		; //* byte align row */
  FAXMODE_WORDALIGN	= $0008		; //* word align row */
  FAXMODE_CLASSF	= FAXMODE_NORTC	; //* TIFF Class F */
  TIFFTAG_JPEGQUALITY		= 65537	; //* Compression quality level */
//* Note: quality level is on the IJG 0-100 scale.  Default value is 75 */
  TIFFTAG_JPEGCOLORMODE		= 65538	; //* Auto RGB<=>YCbCr convert? */
  JPEGCOLORMODE_RAW	= $0000		; //* no conversion (default) */
  JPEGCOLORMODE_RGB	= $0001		; //* do auto conversion */
  TIFFTAG_JPEGTABLESMODE		= 65539	; //* What to put in JPEGTables */
  JPEGTABLESMODE_QUANT = $0001		; //* include quantization tbls */
  JPEGTABLESMODE_HUFF	= $0002		; //* include Huffman tbls */
//* Note: default is JPEGTABLESMODE_QUANT | JPEGTABLESMODE_HUFF */
  TIFFTAG_FAXFILLFUNC		= 65540	; //* G3/G4 fill function */
  TIFFTAG_PIXARLOGDATAFMT		= 65549	; //* PixarLogCodec I/O data sz */
  PIXARLOGDATAFMT_8BIT	= 0	; //* regular u_char samples */
  PIXARLOGDATAFMT_8BITABGR	= 1	; //* ABGR-order u_chars */
  PIXARLOGDATAFMT_11BITLOG	= 2	; //* 11-bit log-encoded (raw) */
  PIXARLOGDATAFMT_12BITPICIO	= 3	; //* as per PICIO (1.0==2048) */
  PIXARLOGDATAFMT_16BIT	= 4	; //* signed short samples */
  PIXARLOGDATAFMT_FLOAT	= 5	; //* IEEE float samples */
//* 65550-65556 are allocated to Oceana Matrix <dev@oceana.com> */
  TIFFTAG_DCSIMAGERTYPE           = 65550   ; //* imager model & filter */
  DCSIMAGERMODEL_M3           = 0       ; //* M3 chip (1280 x 1024) */
  DCSIMAGERMODEL_M5           = 1       ; //* M5 chip (1536 x 1024) */
  DCSIMAGERMODEL_M6           = 2       ; //* M6 chip (3072 x 2048) */
  DCSIMAGERFILTER_IR          = 0       ; //* infrared filter */
  DCSIMAGERFILTER_MONO        = 1       ; //* monochrome filter */
  DCSIMAGERFILTER_CFA         = 2       ; //* color filter array */
  DCSIMAGERFILTER_OTHER       = 3       ; //* other filter */
  TIFFTAG_DCSINTERPMODE           = 65551   ; //* interpolation mode */
  DCSINTERPMODE_NORMAL        = $0     ; //* whole image, default */
  DCSINTERPMODE_PREVIEW       = $1     ; //* preview of image (384x256) */
  TIFFTAG_DCSBALANCEARRAY         = 65552   ; //* color balance values */
  TIFFTAG_DCSCORRECTMATRIX        = 65553   ; //* color correction values */
  TIFFTAG_DCSGAMMA                = 65554   ; //* gamma value */
  TIFFTAG_DCSTOESHOULDERPTS       = 65555   ; //* toe & shoulder points */
  TIFFTAG_DCSCALIBRATIONFD        = 65556   ; //* calibration file desc */
//* Note: quality level is on the ZLIB 1-9 scale. Default value is -1 */
  TIFFTAG_ZIPQUALITY		= 65557	; //* compression quality level */
  TIFFTAG_PIXARLOGQUALITY		= 65558	; //* PixarLog uses same scale */
//* 65559 is allocated to Oceana Matrix <dev@oceana.com> */
  TIFFTAG_DCSCLIPRECTANGLE	= 65559	; //* area of image to acquire */
  TIFFTAG_SGILOGDATAFMT		= 65560	; //* SGILog user data format */
  SGILOGDATAFMT_FLOAT		= 0	; //* IEEE float samples */
  SGILOGDATAFMT_16BIT		= 1	; //* 16-bit samples */
  SGILOGDATAFMT_RAW		= 2	; //* uninterpreted data */
  SGILOGDATAFMT_8BIT		= 3	; //* 8-bit RGB monitor values */
  TIFFTAG_SGILOGENCODE		= 65561 ; //* SGILog data encoding control*/
  SGILOGENCODE_NODITHER	= 0     ; //* do not dither encoded values*/
  SGILOGENCODE_RANDITHER	= 1     ; //* randomly dither encd values */

type
  TRGBQuadArray = array[0..65536] of TRGBQuad;
  PRGBQuadArray = ^TRGBQuadArray;

  TWordArray = array[0..65535] of Word;
  PWordArray = ^TWordArray;

  TByteArray = array[0..1000000] of Byte;
  PByteArray = ^TByteArray;

  PTIFFInfo = ^TTiffInfo;

  TTIFFDirectory = record
	//* bit vector of fields that are set */
    td_fieldsset: array[0..FIELD_SETLONGS-1] of DWORD;
    td_imagewidth, td_imagelength, td_imagedepth: DWORD;
    td_tilewidth, td_tilelength, td_tiledepth: DWORD;
    td_subfiletype: DWORD;
    td_bitspersample: WORD;
    td_sampleformat: WORD;
    td_compression: WORD;
    td_photometric: WORD;
    td_threshholding: WORD;
    td_fillorder: WORD;
    td_orientation: WORD;
    td_samplesperpixel: WORD;
    td_rowsperstrip: DWORD;
    td_minsamplevalue, td_maxsamplevalue: WORD;
    td_sminsamplevalue, td_smaxsamplevalue: Double;
    td_xresolution, td_yresolution: Single;
    td_resolutionunit: WORD;
    td_planarconfig: WORD;
    td_xposition, td_yposition: Single;
    td_pagenumber: array[0..1] of WORD;
    td_colormap: array[0..2] of PWORD;
    td_halftonehints: array[0..1] of WORD;
    td_extrasamples: WORD;
    td_sampleinfo: PWORD;
    td_stonits: Double;
    td_documentname: PChar;
    td_artist: PChar;
    td_datetime: PChar;
    td_hostcomputer: PChar;
    td_imagedescription: PChar;
    td_make: PChar;
    td_model: PChar;
    td_software: PChar;
    td_copyright: PChar;
    td_pagename: PChar;
    td_stripsperimage: DWORD;
    td_nstrips: DWORD;		//* size of offset & bytecount arrays */
    td_stripoffset: PDWORD;
    td_stripbytecount: PDWORD;
//#if SUBIFD_SUPPORT
    td_nsubifd: WORD;
    td_subifd: PDWORD;
//#endif
//#ifdef YCBCR_SUPPORT
    td_ycbcrcoeffs: PSingle;
    td_ycbcrsubsampling: array[0..1] of WORD;
    td_ycbcrpositioning: WORD;
//#endif
//#ifdef COLORIMETRY_SUPPORT
     td_whitepoint: PSingle;
     td_primarychromas: PSingle;
     td_refblackwhite: PSingle;
     td_transferfunction: array[0..2] of PSingle;
//#endif
//#ifdef CMYK_SUPPORT
     td_inkset: WORD;
     td_ninks: WORD;
     td_dotrange: array[0..1] of WORD;
     td_inknameslen: Integer;
     td_inknames: PChar;
     td_targetprinter: PChar;
//#endif
//#ifdef ICC_SUPPORT
     td_profileLength: DWORD;
     td_profileData: Pointer;
//#endif
//#ifdef PHOTOSHOP_SUPPORT
     td_photoshopLength: DWORD;
     td_photoshopData: Pointer;
//#endif
//#ifdef IPTC_SUPPORT
     td_richtiffiptcLength: DWORD;
     td_richtiffiptcData: Pointer;
//#endif
        //* Begin Pixar Tag values. */
     td_imagefullwidth, td_imagefulllength: DWORD;
     td_textureformat: PChar;
     td_wrapmodes: PChar;
     td_fovcot: Single;
     td_matrixWorldToScreen: PSingle;
     td_matrixWorldToCamera: PSingle;
 	//* End Pixar Tag Values. */
  end;
  TTIFFHeader = packed record
    tiff_magic: WORD;	//* magic number (defines byte order) */
    tiff_version: WORD;	//* TIFF version number */
    tiff_diroff: DWORD;	//* byte offset to first directory */
  end;
  TTIFFBoolMethod = function( tiff: PTIFFInfo ): Integer; cdecl;
  TTIFFPreMethod = function( tiff: PTIFFInfo; w: WORD ): Integer; cdecl;
  TTIFFCodeMethod = function( tiff: PTIFFInfo; p: Pointer; u: DWORD; w: WORD ): Integer; cdecl;
  TTIFFVoidMethod = procedure( tiff: PTIFFInfo ); cdecl;
  TTIFFSeekMethod = function( tiff: PTIFFInfo; u: DWORD ): Integer; cdecl;
  TTIFFStripMethod = function( tiff: PTIFFInfo; u: DWORD ): DWORD; cdecl;
  TTIFFTileMethod = procedure( tiff: PTIFFInfo; u, v: DWORD ); cdecl;
  TTIFFPostMethod = procedure( tiff: PTIFFInfo; p: Pointer; n: DWORD ); cdecl;
  TTIFFVSetMethod = function( tiff: PTIFFInfo; u: DWORD; list: Pointer ): Integer; cdecl;
  TTIFFVGetMethod = function( tiff: PTIFFInfo; u: DWORD; list: Pointer ): Integer; cdecl;
  TTIFFPrintMethod = procedure( tiff: PTIFFInfo; f: Pointer; i: Integer ); cdecl;
  TTIFFMapFileProc = function( strm: PStream; pp: PPointer; i: PInteger ): Integer; cdecl;
  TTIFFReadWriteProc = function( strm: PStream; buf: Pointer; sz: DWORD ): DWORD; cdecl;
  TTIFFSeekProc = function( strm: PStream; wh: DWORD; mt: Integer ): DWORD; cdecl;
  TTIFFCloseProc = function( strm: PStream ): Integer; cdecl;
  TTIFFSizeProc = function( strm: PStream ): Integer; cdecl;

  TTiffInfo = record
    tif_name: PChar;	//* name of open file */
    tif_fd: HFile;  	//* open file descriptor */
    tif_mode: Integer;	//* open mode (O_*) */
    tif_flags: DWORD;
    tif_diroff: Integer;	//* file offset of current directory */
    tif_nextdiroff: Integer;	//* file offset of following directory */
    tif_dir: TTIFFDirectory;	//* internal rep of current directory */
    tif_header: TTIFFHeader;	//* file's header block */
    tif_clientdir: Pointer;     //* client TIFF directory */
    tif_typeshift: PInteger;	//* data type shift counts */
    tif_typemask: PInteger;	//* data type masks */
    tif_row: DWORD;	        //* current scanline */
    tif_curdir: WORD {DWORD?};  //* current directory (index) */
    tif_curstrip: DWORD;	//* current strip for read/write */
    tif_curoff: DWORD;	        //* current offset for read/write */
    tif_dataoff: DWORD;	        //* current offset for writing dir */
//#if SUBIFD_SUPPORT
    tif_nsubifd: WORD;          //* remaining subifds to write */
    tif_subifdoff: DWORD;	//* offset for patching SubIFD link */
//#endif
//* tiling support */
    tif_col: DWORD;	        //* current column (offset by row too) */
    tif_curtile: Pointer;	//* current tile for read/write */
    tif_tilesize: DWORD;	//* # of bytes in a tile */
//* compression scheme hooks */
    tif_setupdecode: TTIFFBoolMethod;   //* called once before predecode */
    tif_predecode: TTIFFPreMethod;	//* pre- row/strip/tile decoding */
    tif_setupencode: TTIFFBoolMethod;   //* called once before preencode */
    tif_preencode: TTIFFPreMethod;	//* pre- row/strip/tile encoding */
    tif_postencode: TTIFFBoolMethod;	//* post- row/strip/tile encoding */
    tif_decoderow: TTIFFCodeMethod;	//* scanline decoding routine */
    tif_encoderow: TTIFFCodeMethod;	//* scanline encoding routine */
    tif_decodestrip: TTIFFCodeMethod;   ///* strip decoding routine */
    tif_encodestrip: TTIFFCodeMethod;   //* strip encoding routine */
    tif_decodetile: TTIFFCodeMethod;	//* tile decoding routine */
    tif_encodetile: TTIFFCodeMethod;	//* tile encoding routine */
    tif_close: TTIFFVoidMethod;	        //* cleanup-on-close routine */
    tif_seek: TTIFFSeekMethod;	        //* position within a strip routine */
    tif_cleanup: TTIFFVoidMethod;	//* cleanup state routine */
    tif_defstripsize: TTIFFStripMethod; //* calculate/constrain strip size */
    tif_deftilesize: TTIFFTileMethod;   //* calculate/constrain tile size */
    tif_data: Pointer;	                //* compression scheme private data */
//* input/output buffering */
    tif_scanlinesize: DWORD;            //* # of bytes in a scanline */
    tif_scanlineskew: DWORD;            //* scanline skew for reading strips */
    tif_rawdata: DWORD;	                //* raw data buffer */
    tif_rawdatasize: DWORD;             //* # of bytes in raw data buffer */
    tif_rawcp: Pointer;	                //* current spot in raw buffer */
    tif_rawcc: DWORD;	                //* bytes unread from raw buffer */
//* memory-mapped file support */
    tif_base: Pointer;	                //* base of mapped file */
    tif_size: DWORD;	                //* size of mapped file region (bytes) */
    tif_mapproc: TTIFFMapFileProc;	//* map file method */
    tif_unmapproc: TTIFFMapFileProc;    //* unmap file method */
//* input/output callback methods */
    tif_clientdata: PStream;	        //* callback parameter */
    tif_readproc: TTIFFReadWriteProc;	//* read method */
    tif_writeproc: TTIFFReadWriteProc;  //* write method */
    tif_seekproc: TTIFFSeekProc; 	//* lseek method */
    tif_closeproc: TTIFFCloseProc;	//* close method */
    tif_sizeproc: TTIFFSizeProc;	//* filesize method */
//* post-decoding support */
    tif_postdecode: TTIFFPostMethod;	//* post decoding routine */
//* tag support */
    tif_fieldinfo: Pointer {TIFFFieldInfo**};	//* sorted table of registered tags */
    tif_nfields: Integer;	        //* # entries in registered tag table */
    tif_vsetfield: TTIFFVSetMethod;	//* tag set routine */
    tif_vgetfield: TTIFFVGetMethod;	//* tag get routine */
    tif_printdir: TTIFFPrintMethod;	//* directory print routine */
//* support for read selected tile or strip only */

    tif_read_single_tile_only: DWORD;
    tif_read_tile_xcol: DWORD;
    tif_read_tile_yrow: DWORD;
    tif_jpeg_scale: DWORD; //* 0 - no scale; 1,2,3 - downscale to 2**N times */
  end;


  PTIFF = ^TTIFF;


  TTIFFImageInfo = packed record

    dwEffWidth: Integer;			///< DWORD aligned scan line width

    pImage: PByteArray;				///< THE IMAGE BITS
    pGhost: PTIFF;			///< if this is a ghost, pGhost points to the body
    pParent: PTIFF;			///< if this is a layer, pParent points to the body
    dwType: DWORD;				///< original image format
    szLastError: array[0..255] of Char;	///< debugging
    nProgress: Integer;			///< monitor
    nEscape: Integer;			///< escape
    nBkgndIndex: Integer;		///< used for GIF, PNG, MNG
    nBkgndColor: Integer;		///< used for RGB transparency
    nQuality: BYTE;			///< used for JPEG
    nJpegScale: BYTE;			///< used for JPEG [ignacio]
    nFrame: Integer;				///< used for TIF, GIF, MNG : actual frame
    nNumFrames: Integer;			///< used for TIF, GIF, MNG : total number of frames
    dwFrameDelay: DWORD;		///< used for GIF, MNG
    xDPI: Integer;				///< horizontal resolution
    yDPI: Integer;				///< vertical resolution
    rSelectionBox: TRect;		///< bounding rectangle
    nAlphaMax: BYTE;			///< max opacity (fade)
    bAlphaPaletteEnabled: Boolean; ///< true if alpha values in the palette are enabled.
    bEnabled: Boolean;			///< enables the painting functions
    xOffset: Integer;
    yOffset: Integer;
    dwCodecOpt: array[0..{CMAX_IMAGE_FORMATS-}1] of DWORD; ///< for GIF, TIF : 0=def.1=unc,2=fax3,3=fax4,4=pack,5=jpg
    last_c: TRGBQuad;				///< for GetNearestIndex optimization
    last_c_index: Byte;
    last_c_isvalid: Boolean;
    nNumLayers: Integer;
    dwFlags: DWORD;			///< 0x??00000 = reserved, 0x00??0000 = blend mode, 0x0000???? = layer id - user flags
  end;

  TOnTiffProgress = procedure( Sender: PTIFF; Stage, Progress: Integer; var Cancel: Boolean )
    of object;

  TTIFF = object( TObj )
  private
    fHeight: Integer;
    fWidth: Integer;
    fOnProgress: TOnTiffProgress;
    fReadSingleTile: Boolean;
    fReadXCol: Integer;
    fReadYRow: Integer;
    fJpegScale: TJPEGScale;
    function GetPaletteEntries(idx: Integer): TRGBQuad;
    function GetBpp: Integer;
  protected
    tif: PTiffInfo;
    info: TTIFFImageInfo;
    head: TBitmapInfoHeader;
    pDib: Pointer;
    pAlpha: PByteArray;
    fLoadingFromFile: String;
    samplesperpixel: Integer;
    bitspersample: Integer;
    rowsperstrip: Integer;
    photometric: Integer;
    compression: Integer;
    orientation: Integer;
    function ReadOK( var buf; size: DWORD ): Boolean;
    procedure InitOrder(magic: Integer);
    procedure SetXDPI( dpi: Integer );
    procedure SetYDPI( dpi: Integer );
    function CreateImage(dwWidth, dwHeight: Integer; wBpp: Word; ImageType: DWORD): Pointer;
    function DestroyImage: Boolean;
    function GetSize: DWORD;
    function GetPaletteSize: Integer;
    function GetBits( row: Integer ): Pointer;
    function GetPalette: Pointer;
    procedure SetPalette( pal: PRGBQuadArray; nColors: Integer );
    procedure AlphaCreate;
    procedure AlphaDelete;
    function SetCodecOption( opt, imgtype: Integer ): Boolean;
    function GetField( f: DWORD; valptr: Pointer ): Boolean;
    function GetField3( f: DWORD; valptr1, valptr2, valptr3: Pointer ): Boolean;
    function GetFieldDefaulted( f: DWORD; valptr: Pointer ): Boolean;
    function CalculateLine( w, bdepth: Integer ): Integer;
    function DecodeRGB: Boolean;
    function DecodeWithPalette: Boolean;
    procedure TileToStrip( _out: PByte; _in: PByte; rows, cols: Integer;
              outskew, inskew: Integer );
    procedure SetPixelIndex( x, y: Integer; i: Byte );
    procedure SetPixelColor( x, y: Integer; c: TRGBQUAD; bSetAlpha: Boolean );
    function GetNearestIndex( c: TRGBQUAD ): Byte;
    procedure Mirror;
    procedure Flip;
    procedure RotateRight;
    procedure RotateLeft;
    procedure ApplyAlpha( DestBmp, BackBmp: PBitmap; X, Y: Integer );
  protected
    fNumFrames: Integer;
    fFrame2Decode: Integer;
    fGetDimensionOnly: Boolean;
    rwidth, rheight: Integer;
    fIsJpeg: Boolean;
  public
    destructor Destroy; virtual;
    function LoadFromFile( const F_name: String ): Boolean;
    function LoadFromStream( Strm: PStream ): Boolean;
    procedure Clear;
    property Height: Integer read fHeight;
    property Width: Integer read fWidth;
    function Empty: Boolean;
    procedure Draw( DC: HDC; X, Y: Integer );
    procedure DrawTransparent( DC: HDC; X, Y: Integer; TargetWidth, TargetHeight: Integer );
    procedure StretchDraw( DC: HDC; const R: TRect );
    function PixelFormat: TPixelFormat;
    function PaletteSize: Integer;
    property PaletteEntries[ idx: Integer ]: TRGBQuad read GetPaletteEntries;
    procedure MakeThumbnail( Bmp: PBitmap );
    function ImageAsBitmap: PBitmap;
    function Transparent: Boolean;
    property AlphaMask: PByteArray read pAlpha;
    property Bpp: Integer read GetBpp;
    property OnProgress: TOnTiffProgress read fOnProgress write fOnProgress;

    //---- FRAMES -----
    property Frame: Integer read fFrame2Decode write fFrame2Decode;
    property CountFrames: Integer read fNumFrames; // to obtain CountFrames only
             // w/o reading any frame, set Frame := -1;
    property GetDimensionOnly: Boolean read fGetDimensionOnly write fGetDimensionOnly;

    //---- PARTIAL READ -----
    property ReadSingleTile: Boolean read fReadSingleTile write fReadSingleTile;
    property Tile2Read_nXCol: Integer read fReadXCol write fReadXCol;
    property Tile2Read_nYRow: Integer read fReadYRow write fReadYRow;

    //----- JPEG / OJPEG reducing ----
    property JpegScale: TJPEGScale read fJpegScale write fJpegScale;
             // do not use for images which actually are not jpeg-compressed!
             // If you are not sure, set first GetDimensionOnly to TRUE and
             // perform LoadFromStream, then retrive IsJpegCompressed property.
    property IsJpegCompressed: Boolean read fIsJpeg;
  end;

function NewTIFF: PTIFF;

implementation //###############################################################

uses tif_open, tif_compress, tif_swab, tif_dirread, tif_dir, tif_aux, ex_math,
  tif_getimage, tif_strip, tif_tile, tif_read, tif_close;

function NewTIFF: PTIFF;
begin
  new( Result, Create );
end;

function i2min( x, y: Integer ): Integer;
asm
  cmp  eax, edx
  jle  @@1
  xchg eax, edx
@@1:
end;

function readproc( strm: PStream; buf: PChar; size: DWORD ): Integer; cdecl;
begin
  Result := strm.Read( buf^, size );
end;

function writedummy( strm: PStream; buf: PChar; size: DWORD ): Integer; cdecl;
begin
  Result := 0;
end;

function seekproc( strm: PStream; off: DWORD; mt: Integer ): Integer; cdecl;
begin
  Result := strm.Seek( off, TMoveMethod( mt ) );
end;

function closeproc( strm: PStream ): Integer; cdecl;
begin
  Result := 0;
end;

function sizeproc( strm: PStream ): Integer; cdecl;
begin
  Result := strm.Size;
end;

procedure mapdummy( strm: PStream; base: Pointer; size: DWORD ); cdecl;
begin
end;

procedure unmapdummy( strm: PStream; base: Pointer; size: DWORD ); cdecl;
begin
end;

type T_TIFFSetDefaultCompressionState = procedure( tiff: PTIFFInfo ); cdecl;

{ TTIFF }

procedure TTIFF.AlphaCreate;
begin
  if (pAlpha = nil) then
    pAlpha := AllocMem(head.biWidth * head.biHeight);
end;

procedure TTIFF.AlphaDelete;
begin
  if pAlpha <> nil then FreeMem( pAlpha );
  pAlpha := nil;
end;

procedure TTIFF.ApplyAlpha(DestBmp, BackBmp: PBitmap; X, Y: Integer);
var i, j: Integer;
    AlphaByte: PByte;
    Dest, Back: PRGBQuad;
    A: Byte;
begin
  if pAlpha = nil then Exit;
  if Y < 0 then Y := 0;
  if X < 0 then X := 0;
  for i := Y to Y + DestBmp.Height-1 do
  begin
    if i < Height then
    begin
      AlphaByte := Pointer( Integer( pAlpha ) + (Height - i - 1) * Width + X );
      Dest := DestBmp.ScanLine[ i-Y ];
      Back := BackBmp.ScanLine[ i-Y ];
      for j := 0 to DestBmp.Width-1 do
      begin
        A := AlphaByte^;
        Dest.rgbBlue := (Dest.rgbBlue * A + Back.rgbBlue * (255 - A)) div 255;
        Dest.rgbRed  := (Dest.rgbRed  * A + Back.rgbRed  * (255 - A)) div 255;
        Dest.rgbGreen:= (Dest.rgbGreen* A + Back.rgbGreen* (255 - A)) div 255;
        inc( Dest );
        inc( Back );
        inc( AlphaByte );
      end;
    end;
  end;
end;

function TTIFF.CalculateLine(w, bdepth: Integer): Integer;
begin
  Result := ( (w * bdepth) + 7 ) div 8;
end;

procedure TTIFF.Clear;
begin
  FillChar( head, SizeOf( head ), #0 );
  DestroyImage;
end;

function TTIFF.CreateImage(dwWidth, dwHeight: Integer; wBpp: Word;
  ImageType: DWORD): Pointer;
var pal: PRGBQuad;
begin
(**
 * Initializes or rebuilds the image.
 * \param dwWidth: width
 * \param dwHeight: height
 * \param wBpp: bit per pixel, can be 1, 4, 8, 24
 * \param imagetype: (optional) set the image format, see ENUM_CXIMAGE_FORMATS
 * \return pointer to the internal pDib object; NULL if an error occurs.
 *)
 Result := nil;
	// destroy the existing image (if any)
  if (not DestroyImage) then Exit;

  // prevent further actions if width or height are not vaild <Balabasnia>
  if ((dwWidth = 0) or (dwHeight = 0)) then
  begin
    //strcpy(info.szLastError,'CreateImage : width and height must be greater than zero');
    Exit;
  end;

  // Make sure bits per pixel is valid
  if      (wBpp <= 1) then	wBpp := 1
  else if (wBpp <= 4) then	wBpp := 4
  else if (wBpp <= 8) then	wBpp := 8
  else			        wBpp := 24;

  // limit memory requirements (and also a check for bad parameters)
  if (((dwWidth*dwHeight*wBpp) shr 8) > IMAGE_MAX_MEMORY)then
  begin
    //strcpy(info.szLastError,"IMAGE_MAX_MEMORY exceeded");
    Exit;
  end;


      // set the correct bpp value
  CASE (wBpp) of
  1: head.biClrUsed := 2;
  4: head.biClrUsed := 16;
  8: head.biClrUsed := 256;
  else head.biClrUsed := 0;
  END;

  //set the common image informations
  info.dwEffWidth := ((((wBpp * dwWidth) + 31) div 32) * 4);
  info.dwType := imagetype;

  // initialize BITMAPINFOHEADER
  head.biSize := sizeof(TBITMAPINFOHEADER); //<ralphw>
  head.biWidth := dwWidth;		// fill in width from parameter
  head.biHeight := dwHeight;	// fill in height from parameter
  head.biPlanes := 1;			// must be 1
  head.biBitCount := wBpp;		// from parameter
  head.biCompression := BI_RGB;
  head.biSizeImage := info.dwEffWidth * dwHeight;
//    head.biXPelsPerMeter := 0; See SetXDPI
//    head.biYPelsPerMeter := 0; See SetYDPI
  head.biClrImportant := 0;

  pDib := Pointer(
    GlobalAlloc( GMEM_FIXED {or GMEM_ZEROINIT}, GetSize )
    ); // alloc memory block to store our bitmap
  if (pDib = nil)then
  begin
    //strcpy(info.szLastError,"CreateImage can't allocate memory");
    Exit;
  end;

  //clear the palette
  pal := GetPalette();
  if (pal <> nil) then FillChar( pal^, GetPaletteSize(), #0);

  //Destroy the existing alpha channel
  AlphaDelete();

  // use our bitmap info structure to fill in first part of
  // our DIB with the BITMAPINFOHEADER
  PBitmapInfoHeader( pDib )^ := head;

  info.pImage := GetBits( 0 );
  Result := pDib;
end;

function TTIFF.DecodeRGB: Boolean;
var raster: PDWORD;
    row: PDWORD;
    bits: PByte;  // to source area
    bits2: PByte; // to destination area
    x, y: Integer;
    LinesPerProgress, NextLineProgress: Integer;
    Cancel: Boolean;
begin
  // Read the whole image into one big RGBA buffer using
  // the traditional TIFFReadRGBAImage() API that we trust.
  // retrieve RGBA image
  Result := FALSE;

  GetMem( raster, fWidth * fHeight * sizeof(DWORD) );
  TRY

    if (raster = nil) then
    begin
      //throw "No space for raster buffer";
      Exit;
    end;

    // Read the image in one chunk into an RGBA array
    if(j__TIFFReadRGBAImage(Integer(tif), rwidth, rheight, Integer( raster ), 1) = 0) then
      Exit;

    // read the raster lines and save them in the DIB
    // with RGB mode, we have to change the order of the 3 samples RGB
    row := raster;
    bits2 := Pointer( info.pImage );

    LinesPerProgress := max( 1, 1000000 div (Width * 3 + 1) );
    NextLineProgress := 0;

    for y := 0 to fHeight-1 do
    begin

      if (y >= NextLineProgress) and Assigned( fOnProgress ) then
      begin
        inc( NextLineProgress, LinesPerProgress );
        Cancel := FALSE;
        fOnProgress( @ Self, 1, (y+1) * 100 div Height, Cancel );
        if Cancel then Exit;
      end;

      if (info.nEscape <> 0) then  // <vho> - cancel decoding
      begin
        //throw "Cancelled";
        Exit;
      end;

      bits := bits2;
      for x := 0 to Width-1 do
      begin
        bits^ := row^ shr 16; inc( bits ); // B
        bits^ := row^ shr 8 ; inc( bits ); // G
        bits^ := row^       ; inc( bits ); // R
        if (samplesperpixel = 4) then
          PByte( Integer( pAlpha ) + Width * y + x )^ := row^ shr 24;
        inc( row );
      end;
      inc( bits2, info.dwEffWidth );
    end;

    Result := TRUE;

  FINALLY
    if raster <> nil then FreeMem( raster );
  END;
end;

procedure MoveBits( From: PByte; Dest: PByte; count, bpp: Integer );
var offbits: Integer;
    w: Word;
    d: DWORD;
begin
  offbits := 0;
  if bpp <= 8 then
    while count > 0 do
    begin
      if offbits + bpp <= 8 then
        w := From^ shr (8 - offbits - bpp)
      else
      begin
        w := From^ shl (offbits + bpp - 8);
        inc( From );
        w := w or (From^ shr (16 - offbits - bpp));
      end;
      inc( offbits, bpp );
      if offbits >= 8 then
      begin
        dec( offbits, 8 );
        if offbits = 0 then
          inc( From );
      end;
      if bpp > 8 then
        w := w shr (bpp - 8)
      else
      if bpp < 8 then
        w := w and ((1 shl bpp)-1);
      Dest^ := w;
      inc( Dest );
      dec( Count );
    end
  else if bpp < 16 then
    while count > 0 do
    begin
      d := From^ shl 24 or PByte( Integer( From ) + 1 )^ shl 16 or
                           PByte( Integer( From ) + 2 )^ shl 8 or
                           PByte( Integer( From ) + 3 )^;
      Dest^ := d shr (32 - offbits {- bpp + bpp} - 8);
      inc( Dest );
      inc( offbits, bpp );
      while offbits >= 8 do
      begin
        inc( From );
        dec( offbits, 8 );
      end;
      dec( count );
    end
  else if bpp < 32 then
    while count > 0 do
    begin
      d := From^ shl 24 or PByte( Integer( From ) + 1 )^ shl 16 or
                           PByte( Integer( From ) + 2 )^ shl 8 or
                           PByte( Integer( From ) + 3 )^;
      Dest^ := d shr (offbits + bpp - 8);
      inc( Dest );
      inc( offbits, bpp );
      while offbits >= 8 do
      begin
        inc( From );
        dec( offbits, 8 );
      end;
      dec( count );
    end
  else
    while count > 0 do
    begin
      d := PDWORD(From)^;
      Dest^ := d shr 24;
      inc( Dest );
      inc( From, 4 );
      dec( count );
    end
end;

procedure MoveBitsPal( from: PByteArray; dest: PByte; count, bpp: Integer;
          const pal: PRGBQuadArray );
var offbits: Integer;
    d: DWORD;
    palidx: WORD;
begin
  offbits := 0;
  while (count > 0) do
  begin
    d := (from[0] shl 24) or ( from[1] shl 16 )
     or (from[2] shl 8 ) or (from[3]);
    palidx := (d shr (32 - offbits - bpp));
    if (bpp < 16) then
    begin
      palidx := palidx shl (16-bpp);
      palidx := (palidx shr 8) or (palidx shl 8);
      palidx := palidx shr (16-bpp);
    end
      else palidx := (palidx shr 8) or (palidx shl 8);
    dest^ := pal[palidx].rgbBlue;  inc( dest );
    dest^ := pal[palidx].rgbGreen; inc( dest );
    dest^ := pal[palidx].rgbRed;   inc( dest );
    inc( offbits, bpp );
    while (offbits >= 8) do
    begin
      from := Pointer( Integer( from ) + 1 );
      dec( offbits, 8 );
    end;
    dec( count );
  end;
end;


function TTIFF.DecodeWithPalette: Boolean;
var pal: PRGBQuadArray;
    red, green, blue: PWordArray;
    Palette16Bits: Boolean;
    nrow, ys: Integer;
    line, line2, bitsize: Integer;
    tiled_image: Boolean;
    tw, tl: Integer;
    tilebuf: PByte;
    imagew, tilew: Integer;
    iskew: Integer;
    bufp: PByteArray;
    colb, col: Integer;
    owidth, oskew: Integer;
    xi, yi, ii, i, n, x, y: Integer;
    c: TRGBQuad;
    l,a,b,bitsoffset: Integer;
    p,cx,cy,cz,cr,cg,cb: Extended;
    bits: PByteArray;  // to source area
    oldbits: PByteArray;  // to source area
    bits2: PByte; // to destination area
    offset, offset2: Integer;
    _bpp: Integer;
    BIG_palette: Boolean;
    reversed: Integer;
    unite0, unite1: Byte;
    {$IFDEF FIX_16BPP_DARKIMG}
    row_shifts: PByteArray;
    the_shift: Integer;
    hi_max, hi_byte: Byte;
    data: PByte;
    need_shift: Integer;
    min_row_shift: Byte;
    {$ENDIF}

    LinesPerProgress: Integer;
    NextLineProgress: Integer;
    Cancel: Boolean;

    n_xcol, n_yrow: Integer;
    tile_ready, tile_bad: Boolean;
begin
  Result := FALSE;
  BIG_palette := (bitspersample > 8) and	// + VK
                 (photometric = PHOTOMETRIC_PALETTE);
  if BIG_palette and (bitspersample > 24) then	// + VK
    Exit;

  _bpp := bitspersample;
  if _bpp > 8 then _bpp := 8;

  if BIG_palette then
    pal := AllocMem( (1 shl bitspersample) * sizeof(TRGBQUAD) )
  else
    pal := AllocMem( 256 * sizeof(TRGBQUAD) );
  if (pal = nil) then
  begin
    //throw "Unable to allocate TIFF palette";
    Exit;
  end;

  TRY
    // set up the colormap based on photometric
    CASE(photometric) OF
    PHOTOMETRIC_MINISBLACK,	// bitmap and greyscale image types
    PHOTOMETRIC_MINISWHITE:
      if (bitspersample = 1) then // Monochrome image
      begin
        if (photometric = PHOTOMETRIC_MINISBLACK) then
          DWORD( pal[1] ) := $FFFFFF
        else
          DWORD( pal[0] ) := $FFFFFF;
      end
        else // need to build the scale for greyscale images
      begin
        if (photometric = PHOTOMETRIC_MINISBLACK) then
          for i:=0 to (1 shl _bpp)-1 do
          begin
            b := i*(255 div ((1 shl _bpp)-1));
            pal[i].rgbRed   := b;
            pal[i].rgbGreen := b;
            pal[i].rgbBlue  := b;
          end
        else
          for i:=0 to (1 shl _bpp)-1 do
          begin
            b := 255-i*(255 div ((1 shl _bpp)-1));
            pal[i].rgbRed   := b;
            pal[i].rgbGreen := b;
            pal[i].rgbBlue  := b;
          end;
      end;
    PHOTOMETRIC_PALETTE:	// color map indexed
      begin
        if GetField3( TIFFTAG_COLORMAP, @ red, @ green, @ blue ) then
        begin
          // Is the palette 16 or 8 bits ?
          Palette16Bits := BIG_palette;
          if not BIG_palette then
          begin
            n := 1 shl _bpp;
            while (n > 0) do
            begin
              if ((red[n] >= 256) or (green[n] >= 256) or (blue[n] >= 256)) then
              begin
                Palette16Bits := TRUE;
                break;
              end;
              dec( n );
            end;
          end;

          // load the palette in the DIB
          if BIG_palette then i := (1 shl bitspersample) - 1
          else i := (1 shl _bpp) - 1;
          for i := i downto 0 do
          begin
            if (Palette16Bits) then
            begin
              pal[i].rgbRed   := red[i]  * 255 div 65535;
              pal[i].rgbGreen := green[i]* 255 div 65535;
              pal[i].rgbBlue  := blue[i] * 255 div 65535;
            end
              else
            begin
              pal[i].rgbRed   := red[i];
              pal[i].rgbGreen := green[i];
              pal[i].rgbBlue  := blue[i];
            end;
          end;
        end;

      end;
    END;

    if not BIG_palette then
    begin
      SetPalette( pal, 1 shl _bpp );
      FreeMem( pal );
      pal := nil;
    end;

    // read the tiff lines and save them in the DIB
    line := CalculateLine( fWidth, bitspersample * samplesperpixel );
    tiled_image := j__TIFFIsTiled(Integer(tif)) <> 0;
    tilebuf := nil;
    {bitsize := j__TIFFStripSize( Integer(tif) );
    //verify bitsize: could be wrong if StripByteCounts is missing.
    if bitsize > Integer( head.biSizeImage ) * samplesperpixel then
      bitsize := Integer( head.biSizeImage ) * samplesperpixel;}
    bitsize := Integer( head.biSizeImage )*samplesperpixel;
    if (bitspersample > 8) and
       not( (bitspersample = 16) {and tiled_image} ) then
      bitsize := bitsize * ((bitspersample + 7) div 8);

    tw := 0; tl := 0;
    if (tiled_image) then
    begin
      GetField( TIFFTAG_TILEWIDTH, @ tw );
      GetField( TIFFTAG_TILELENGTH, @ tl );
      rowsperstrip := tl;
      bitsize := j__TIFFTileSize(Integer(tif)) * (1 + Width div Integer( tw ));
      GetMem( tilebuf, j__TIFFTileSize(Integer(tif)) );
    end;

    if bitspersample = 16 then
      GetMem( bits, bitsize * 2 )
    else
      GetMem( bits, bitsize );

    oldbits := nil;
    line2 := line;
    if not tiled_image and (bitspersample = 16) then
    begin
      line2 := line;
      line := CalculateLine( Width, 8 * samplesperpixel );
      oldbits := bits;
      GetMem( bits, bitsize );
    end;

    row_shifts := nil;
    TRY
      if bits = nil then Exit;

      {$IFDEF FIX_16BPP_DARKIMG}
      if oldbits <> nil then
        row_shifts := AllocMem( fHeight );
      {$ENDIF}

      LinesPerProgress := max( 1, 1000000 div (line+1) );
      NextLineProgress := 0;

      ys := 0;
      n_yrow := 0;
      while Integer( ys ) < rheight do
      begin

        tile_ready := FALSE;

        if (ys >= NextLineProgress) and Assigned( fOnProgress ) then
        begin
          inc( NextLineProgress, LinesPerProgress );
          Cancel := FALSE;
          fOnProgress( @ Self, 1, (ys+1) * 100 div Height, Cancel );
          if Cancel then Exit;
        end;

        if (info.nEscape <> 0) then // <vho> - cancel decoding
        begin
          //throw "Cancelled";
          Exit;
        end;

        if rowsperstrip < 0 then
          rowsperstrip := rheight;
        if DWORD( ys + rowsperstrip ) > DWORD( rheight ) then
          nrow := rheight - ys
        else
          nrow := rowsperstrip;

        tile_bad := FALSE;
        if (tiled_image) then
        begin
          imagew := j__TIFFScanlineSize(Integer(tif));
          tilew  := j__TIFFTileRowSize (Integer(tif));
          iskew  := imagew - tilew;
          bufp   := Pointer(bits);

          colb   := 0;
          col    := 0;
          n_xcol := 0;
          while col < fWidth do
          begin

            if not fReadSingleTile or
               (fReadXCol = n_xcol) and (fReadYRow = n_yrow) then
            begin
              tile_bad := FALSE;

              if (j__TIFFReadTile(Integer(tif), Integer( tilebuf ), col, ys, 0, 0) < 0) then
              begin
                //throw "Corrupted tiled TIFF file!";
                tile_bad := TRUE;
                if n_yrow = 0 then
                  Exit;
              end;

              if not tile_bad then
              begin
                if (colb + tw > imagew) then
                begin
                  owidth := imagew - colb;
                  oskew  := tilew - owidth;
                  TileToStrip(@bufp[colb], tilebuf, nrow, owidth, oskew + iskew, oskew );
                end
                  else
                begin
                  TileToStrip(@bufp[colb], tilebuf, nrow, tilew, iskew, 0);
                end;
              end;
              inc( colb, tilew );
              tile_ready := TRUE;
              if fReadSingleTile then
                break;
            end;
            inc( col, tw );
            inc( n_xcol );
          end; // while col
        end
        else // not tiled_image
        begin
          if not fReadSingleTile or
             (fReadXCol = 0) and (fReadYRow = n_yrow) then
          begin
            if bitspersample = 16 then
            begin
              if (j__TIFFReadEncodedStrip( Integer(tif),
                                         j__TIFFComputeStrip( Integer(tif), ys, 0),
                                         Integer( oldbits ), nrow * line2) = -1) then
              begin
                if n_yrow = 0 then
                  Exit;
              end;
            end
              else
            begin
                if (j__TIFFReadEncodedStrip( Integer(tif),
                                           j__TIFFComputeStrip( Integer(tif), ys, 0),
                                           Integer( bits ), nrow * line) = -1) then
                begin
                  //throw "Corrupted TIFF file!";
                  tile_bad := TRUE;
                  if n_yrow = 0 then
                    Exit;
                end;
            end;
            tile_ready := TRUE;
          end;
        end; // not tiled_image

        if tile_ready then
        BEGIN
          for y := 0 to nrow-1 do
          begin

            if (ys+y >= NextLineProgress) and Assigned( fOnProgress ) then
            begin
              inc( NextLineProgress, LinesPerProgress );
              Cancel := FALSE;
              fOnProgress( @ Self, 1, (ys+y+1) * 100 div Height, Cancel );
              if Cancel then Exit;
            end;

            offset := (Integer( nrow )-y-1)*line;
            if (bitspersample = 16) and not BIG_palette then
            begin
              offset2 := (Integer( nrow )-y-1)*line2;
              if oldbits <> nil then
              begin
                {$IFDEF FIX_16BPP_DARKIMG}
                hi_max := 0;
                for xi := 0 to line-1 do
                begin
                  hi_byte := oldbits[xi*2+offset2+1];
                  if hi_byte > hi_max then hi_max := hi_byte;
                end;
                the_shift := 0;
                if hi_max = 0 then the_shift := 8
                else
                while( hi_max and $80 = 0 ) do
                begin
                  inc( the_shift );
                  hi_max := hi_max shl 1;
                end;
                row_shifts[height-ys-nrow+y] := the_shift;
                the_shift := 8 - the_shift;
                for xi :=0 to line-1 do
                  bits[xi+offset] := ((oldbits[xi*2+offset2+1]shl 8) or
                                       oldbits[xi*2+offset2]) shr the_shift;
                {$ELSE}
                for xi :=0 to line-1 do
                  bits[xi+offset] := oldbits[xi*2+offset2+1];
                {$ENDIF}
              end
              else
                for xi :=0 to Fwidth-1 do
                  bits[xi+offset] := bits[xi*2+offset+1];
            end;
            if (samplesperpixel = 1) then
            begin
              if BIG_palette then
              begin
                if oldbits <> nil then
                begin
                  offset2 := (nrow-y-1)*line2;
                  if rheight = fHeight then
                    MoveBitsPal( Pointer( Integer( oldbits ) + offset2 ),
                            @ info.pImage[ info.dwEffWidth * (fheight-ys-nrow+y) ],
                            FWidth, bitspersample, pal )
                  else
                    MoveBitsPal( Pointer( Integer( oldbits ) + offset2 ),
                            @ info.pImage[ info.dwEffWidth * (fheight-nrow+y) ],
                            FWidth, bitspersample, pal );
                end
                  else
                begin
                  if rheight = fHeight then
                    MoveBitsPal( Pointer( Integer( bits ) + offset ),
                            @ info.pImage[ info.dwEffWidth * (fheight-ys-nrow+y) ],
                            FWidth, bitspersample, pal )
                  else
                    MoveBitsPal( Pointer( Integer( bits ) + offset ),
                            @ info.pImage[ info.dwEffWidth * (fheight-nrow+y) ],
                            FWidth, bitspersample, pal );
                end;
              end
                else
              if (bitspersample = head.biBitCount) or (bitspersample = 16) then
                //simple 8bpp image
                if rheight = fHeight then
                  move( Pointer( Integer( bits ) + offset )^,
                        info.pImage[ info.dwEffWidth * (height-ys-nrow+y) ],
                        info.dwEffWidth )
                else
                  move( Pointer( Integer( bits ) + offset )^,
                        info.pImage[ info.dwEffWidth * (height-nrow+y) ],
                        info.dwEffWidth )
              else
              begin
                if rheight = fHeight then
                  MoveBits( Pointer( Integer( bits ) + offset ),
                            @ info.pImage[ info.dwEffWidth * (height-ys-nrow+y) ],
                            Width, bitspersample )
                else
                  MoveBits( Pointer( Integer( bits ) + offset ),
                            @ info.pImage[ info.dwEffWidth * (height-nrow+y) ],
                            Width, bitspersample );
              end;
            end
            else if (samplesperpixel = 2) then //8bpp image with alpha layer
            begin
              if pAlpha = nil then AlphaCreate;
              xi := 0;
              ii := 0;
              yi := height-ys-nrow+y;
              if (rheight <> fHeight) then
                yi := height-nrow+y;
              while (ii<line) do
              begin
                SetPixelIndex(xi,yi,bits[ii+offset]);
                //AlphaSet(xi,yi,bits[ii+offset+1]);
                pAlpha[xi+yi*Width] := bits[ii+offset+1];
                inc( ii, 2 );
                inc( xi );
                if (xi >= Width) then
                begin
                  dec( yi );
                  xi := 0;
                end;
              end; // while ii<line
            end
            else //photometric==PHOTOMETRIC_CIELAB
            begin
              if (head.biBitCount<>24) then //fix image
              begin
                CreateImage(Width,Height,24,{IMAGE_FORMAT_TIF} 0);
                if (samplesperpixel = 4) then AlphaCreate();
              end;

              xi := 0;
              ii := 0;
              yi := height-ys-nrow+y;
              while (ii<{line}width) do
              begin
                bitsoffset := ii*samplesperpixel+offset;
                l := bits[bitsoffset];
                a := bits[bitsoffset+1];
                b := bits[bitsoffset+2];
                if (a>127) then dec( a, 256 );
                if (b>127) then dec( b, 256 );
                // lab to xyz
                p := (l/2.55 + 16) / 116.0;
                cx := Power( p + a * 0.002, 3);
                cy := Power( p, 3);
                cz := Power( p - b * 0.005, 3);
                // white point
                cx := cx * 0.95047;
                //cy*=1.000;
                cz := cz * 1.0883;
                // xyz to rgb
                cr :=  3.240479 * cx - 1.537150 * cy - 0.498535 * cz;
                cg := -0.969256 * cx + 1.875992 * cy + 0.041556 * cz;
                cb :=  0.055648 * cx - 0.204043 * cy + 1.057311 * cz;

                if ( cr > 0.00304 ) then
                     cr := 1.055 * Power(cr,0.41667) - 0.055
                else cr := 12.92 * cr;
                if ( cg > 0.00304 ) then
                     cg := 1.055 * Power(cg,0.41667) - 0.055
                else cg := 12.92 * cg;
                if ( cb > 0.00304 ) then
                     cb := 1.055 * Power(cb,0.41667) - 0.055
                else cb := 12.92 * cb;

                c.rgbRed   := max(0,i2min(255,Trunc( cr*255 )));
                c.rgbGreen := max(0,i2min(255,Trunc( cg*255 )));
                c.rgbBlue  := max(0,i2min(255,Trunc( cb*255 )));

                SetPixelColor(xi,yi,{Color2RGBQuad( TColor( c )}c, FALSE);
                if (samplesperpixel = 4) then
                  //AlphaSet(xi,yi,bits[bitsoffset+3]);
                  pAlpha[xi+yi*Width] := bits[bitsoffset+3];
                inc( ii );
                inc( xi );
                if (xi >= Width) then
                begin
                  dec( yi );
                  xi := 0;
                end;
              end; // while ii<line
            end; // photometric==PHOTOMETRIC_CIELAB
          end; // for y
          if fReadSingleTile then
            break;
        END;

        inc( ys, rowsperstrip );
        inc( n_yrow );
      end; // while ys

      {$IFDEF FIX_16BPP_DARKIMG}
      if (row_shifts <> nil) and (samplesperpixel = 1) and
         (bitspersample = 16) and not BIG_palette then
      begin
        // 1. calculate maximum necessary shift
        min_row_shift := 8;
        for y := 0 to fheight-1 do
        begin
          if (min_row_shift > row_shifts[y]) then
            min_row_shift := row_shifts[y];
        end;
        // 2. for rows having less shift value, correct such rows:
        NextLineProgress := 0;
        for y := 0 to height-1 do
        begin

          if (y >= NextLineProgress) and Assigned( fOnProgress ) then
          begin
            inc( NextLineProgress, LinesPerProgress );
            Cancel := FALSE;
            fOnProgress( @ Self, 2, (y+1) * 100 div Height, Cancel );
            if Cancel then Exit;
          end;

          if (min_row_shift < row_shifts[y]) then
          begin
            need_shift := row_shifts[y] - min_row_shift;
            data := @ info.pImage[ info.dwEffWidth * y ];
            for x := 0 to width-1 do
            begin
              data^ := data^ shr need_shift;
              inc( data );
            end;
          end;
        end;
      end;
      {$ENDIF}

      CASE (orientation) OF
      ORIENTATION_TOPRIGHT: //* row 0 top, col 0 rhs */
              Mirror;
      ORIENTATION_BOTRIGHT: //* row 0 bottom, col 0 rhs */
        begin
              Flip;
              Mirror;
        end;
      ORIENTATION_BOTLEFT: //* row 0 bottom, col 0 lhs */
              Flip;
      ORIENTATION_LEFTTOP: //* row 0 lhs, col 0 top */
        begin
              RotateRight;
              Mirror;
        end;
      ORIENTATION_RIGHTTOP: //* row 0 rhs, col 0 top */
              RotateLeft;
      ORIENTATION_RIGHTBOT: //* row 0 rhs, col 0 bottom */
        begin
              RotateLeft;
              Mirror;
        end;
      ORIENTATION_LEFTBOT: //* row 0 lhs, col 0 bottom */
              RotateRight;
      END;

      Result := TRUE;
    FINALLY
      if bits <> nil then
        FreeMem( bits );
      if tiled_image then
        FreeMem( tilebuf );
      if oldbits <> nil then
        FreeMem( oldbits );
      if pal <> nil then
        FreeMem( pal );
      pal := nil;
      if row_shifts <> nil then
        FreeMem( row_shifts );
    END;

  FINALLY
    if pal <> nil then
      FreeMem( pal );
  END;

end;

destructor TTIFF.Destroy;
begin
  Clear;
  inherited;
end;

function TTIFF.DestroyImage: Boolean;
begin
  Result := false;
  //free this only if it's valid and it's not a ghost
  if (info.pGhost = nil) then
  begin
    {if (pLayers <> nil) then
    begin
      for n:=0 to info.nNumLayers-1 do pLayers[n].Free;
      Free_And_Nil( pLayers );
    end;}
    //if (pSelection <> nil) then Free_And_Nil(pSelection);
    if (pAlpha <> nil) then AlphaDelete;
    if (pDib <> nil) then
    begin
      GlobalFree( DWORD( pDib ) );
      pDib := nil;
    end;
    Result := true;
  end;
end;

procedure TTIFF.Draw(DC: HDC; X, Y: Integer);
begin
  if Empty then Exit;
  if SetDIBitsToDevice( DC, X, Y, Width, Height, 0, 0,
    0, Height, GetBits( 0 ), PBitmapInfo( pDib )^, DIB_RGB_COLORS ) <> Height then
  begin
    asm
      nop
    end;
  end;
end;

procedure TTIFF.DrawTransparent(DC: HDC; X, Y: Integer; TargetWidth, TargetHeight: Integer);
var TmpBackground, TmpLayer: PBitmap;
    h: Integer;
    srcY, srcX, X1: Integer;
begin
  if not Transparent then
  begin
    Draw( DC, X, Y );
    Exit;
  end;
  if Empty then Exit;
  if TargetWidth = 0 then
  begin
    TargetWidth := Width;
    if X < 0 then inc( TargetWidth, X );
  end;
  if TargetHeight = 0 then
  begin
    TargetHeight := Height;
    if Y < 0 then inc( TargetHeight, Y );
  end;
  if TargetWidth > Width then TargetWidth := Width;
  if TargetHeight > Height then TargetHeight := Height;
  if (TargetWidth <= 0) or (TargetHeight <= 0) then Exit;
  h := 1;
  while (h < TargetHeight) and ((h + 1)*TargetWidth < 1000000) do
    inc( h );
  TmpBackground := NewDIBBitmap( TargetWidth, h, pf32bit );
  TmpLayer      := NewDIBBitmap( TargetWidth, h, pf32bit );
  TRY
    srcY := 0;
    srcX := 0;
    X1 := X;
    if X < 0 then begin
      srcX := -X;
      X1 := 0;
      X := 0;  if X = 0 then;
    end;
    while Y < TargetHeight do
    begin
      if Y + h > 0 then
      begin
        //        :
        BitBlt( TmpBackground.Canvas.Handle, 0, 0, TmpBackground.Width, TmpBackground.Height,
          DC, X, Y, SRCCOPY );
        // 
        Draw( TmpLayer.Canvas.Handle, -srcX, -srcY );
        ApplyAlpha( TmpLayer, TmpBackground, srcX, srcY );
        TmpLayer.Draw( DC, X1, Y );
      end;
      inc( y, h );
      inc( srcY, h );
    end;
  FINALLY
    TmpBackground.Free;
    TmpLayer     .Free;
  END;
end;

function TTIFF.Empty: Boolean;
begin
  Result := (pDib = nil) or (Width <= 0) or (Height <= 0);
end;

procedure TTIFF.Flip;
begin
  //todo
end;

function TTIFF.GetBits( row: Integer ): Pointer;
begin
  Result := nil;
  if (pDib <> nil) then
  begin
    if (row<head.biHeight) then
      Result := PByte(Integer( pDib ) + PInteger( pDib )^ +
             GetPaletteSize + (info.dwEffWidth * row))
    else
      Result := nil;
  end;
end;

function TTIFF.GetBpp: Integer;
begin
  Result := bitspersample * samplesperpixel;
end;

function TTIFF.GetField(f: DWORD; valptr: Pointer): Boolean;
begin
  Result := j__TIFFGetField( Integer(tif), f, valptr ) <> 0;
end;

function TTIFF.GetField3(f: DWORD; valptr1, valptr2,
  valptr3: Pointer): Boolean;
begin
  asm
    push [valptr3]
    push [valptr2]
    push [valptr1]
    push [f]
    mov  edx, [eax].tif
    push edx
    call _TIFFGetField
    test eax, eax
    setnz al
    mov  [Result], al
  end;
end;

function TTIFF.GetFieldDefaulted(f: DWORD; valptr: Pointer): Boolean;
begin
  Result := j__TIFFGetFieldDefaulted( Integer(tif), f, valptr ) <> 0;
end;

function TTIFF.GetNearestIndex(c: TRGBQUAD): Byte;
var iDst: PByteArray;
    distance, i, j, k, l, m: Integer;
begin
  Result := 0;
  if (pDib = nil) or (head.biClrUsed = 0) then Exit;

  // <RJ> check matching with the previous result
  if info.last_c_isvalid and (DWORD( info.last_c ) = DWORD( c )) then
    Result := info.last_c_index
  else
  begin
    info.last_c := c;
    info.last_c_isvalid := true;

    iDst := Pointer( Integer( pDib ) + sizeof(TBITMAPINFOHEADER) );
    distance := 200000;
    j := 0;
    if head.biClrImportant = 0 then
      m := head.biClrUsed
    else
      m := head.biClrImportant;

    l := 0;
    for i := 0 to m-1 do
    begin
      k := (iDst[l] - c.rgbBlue)*(iDst[l] - c.rgbBlue)+
           (iDst[l+1] - c.rgbGreen)*(iDst[l+1] - c.rgbGreen)+
           (iDst[l+2] - c.rgbRed)*(iDst[l+2] - c.rgbRed);
      // k = abs(iDst[l]-c.rgbBlue)+abs(iDst[l+1]-c.rgbGreen)+abs(iDst[l+2]-c.rgbRed);
      if (k=0) then
      begin
        j := i;
        break;
      end;
      if (k < distance) then
      begin
        distance := k;
        j := i;
      end;
      inc( l, sizeof(RGBQUAD) );
    end;
    info.last_c_index := j;
    Result := j;
  end;
end;

function TTIFF.GetPalette: Pointer;
begin
  Result := nil;
  if ((pDib <> nil)and(head.biClrUsed <> 0)) then
    Result := Pointer(DWORD( pDib ) + sizeof(TBITMAPINFOHEADER));
end;

function TTIFF.GetPaletteEntries(idx: Integer): TRGBQuad;
var pal: PRGBQuad;
    C: TRGBQuad;
begin
  pal := GetPalette;
  DWORD( C ) := 0;
  if pal = nil then
    Result := C
  else
    Result := PRGBQuad( Integer( pal ) + Sizeof( TRGBQuad ) * idx )^;
end;

function TTIFF.GetPaletteSize: Integer;
begin
  Result := head.biClrUsed * sizeof(RGBQUAD);
end;

function TTIFF.GetSize: DWORD;
begin
  Result := head.biSize + head.biSizeImage + DWORD( GetPaletteSize() );
end;

type
  PCrackBitmap = ^TCrackBitmap;
  TCrackBitmap = object( KOL.TBitmap )
  end;

function TTIFF.ImageAsBitmap: PBitmap;
var pf: TPixelFormat;
begin
  Result := nil;
  if Empty then Exit;
  CASE head.biPlanes * head.biBitCount OF
  1: pf := pf1bit;
  4: pf := pf4bit;
  8: pf := pf8bit;
  16: pf := pf16bit;
  24: pf := pf24bit;
  32: pf := pf32bit;
  else Exit;
  END;
  Result := NewDibBitmap( 1, 1, pf );
  GlobalFree( DWORD( Result.DIBBits ) );
  PCrackBitmap( Result ).fWidth := Width;
  PCrackBitmap( Result ).fHeight := Height;
  PCrackBitmap( Result ).fDIBBits := GetBits( 0 );
  PCrackBitmap( Result ).fDIBAutoFree := TRUE;
  move( head, PCrackBitmap( Result ).fDIBHeader.bmiHeader, Sizeof( head ) );
  if GetPalette <> nil then
    move( GetPalette^, PCrackBitmap( Result ).fDibHeader.bmiColors[0],
      GetPaletteSize );
  PCrackBitmap( Result ).fScanLineSize :=
    CalcScanLineSize( @PCrackBitmap( Result ).fDIBHeader.bmiHeader );
end;

procedure TTIFF.InitOrder(magic: Integer);
const
  typemask: array[0..12] of DWORD = (
	0,		//* TIFF_NOTYPE */
	$000000ff,	//* TIFF_BYTE */
	$ffffffff,	//* TIFF_ASCII */
	$0000ffff,	//* TIFF_SHORT */
	$ffffffff,	//* TIFF_LONG */
	$ffffffff,	//* TIFF_RATIONAL */
	$000000ff,	//* TIFF_SBYTE */
	$000000ff,	//* TIFF_UNDEFINED */
	$0000ffff,	//* TIFF_SSHORT */
	$ffffffff,	//* TIFF_SLONG */
	$ffffffff,	//* TIFF_SRATIONAL */
	$ffffffff,	//* TIFF_FLOAT */
	$ffffffff	//* TIFF_DOUBLE */
  );
  bigTypeshift: array[0..12] of Integer = (
	0,		//* TIFF_NOTYPE */
	24,		//* TIFF_BYTE */
	0,		//* TIFF_ASCII */
	16,		//* TIFF_SHORT */
	0,		//* TIFF_LONG */
	0,		//* TIFF_RATIONAL */
	24,		//* TIFF_SBYTE */
	24,		//* TIFF_UNDEFINED */
	16,		//* TIFF_SSHORT */
	0,		//* TIFF_SLONG */
	0,		//* TIFF_SRATIONAL */
	0,		//* TIFF_FLOAT */
	0		//* TIFF_DOUBLE */
  );
  litTypeshift: array[0..12] of Integer = (
	0,		//* TIFF_NOTYPE */
	0,		//* TIFF_BYTE */
	0,		//* TIFF_ASCII */
	0,		//* TIFF_SHORT */
	0,		//* TIFF_LONG */
	0,		//* TIFF_RATIONAL */
	0,		//* TIFF_SBYTE */
	0,		//* TIFF_UNDEFINED */
	0,		//* TIFF_SSHORT */
	0,		//* TIFF_SLONG */
	0,		//* TIFF_SRATIONAL */
	0,		//* TIFF_FLOAT */
	0		//* TIFF_DOUBLE */
  );
begin
  tif.tif_typemask := @ typemask[ 0 ];
  if (magic = TIFF_BIGENDIAN) then
  begin
    tif.tif_typeshift := @ bigTypeshift[ 0 ];
    if (not bigendian) then
            tif.tif_flags := tif.tif_flags or TIFF_SWAB;
  end
    else
  begin
    tif.tif_typeshift := @ litTypeshift[ 0 ];
    if (bigendian) then
            tif.tif_flags := tif.tif_flags or TIFF_SWAB;
  end;
end;

function TTIFF.LoadFromFile(const F_name: String): Boolean;
var Strm: PStream;
begin
  Strm := NewReadFileStream( F_name );
  fLoadingFromFile := F_Name;
  TRY
    Result := LoadFromStream( Strm );
  FINALLY
    Strm.Free;
    fLoadingFromFile := '';
  END;
end;

function TTIFF.LoadFromStream(Strm: PStream): Boolean;
var res_unit: Integer;
    resolution, offset: Single;
    isRGB: Boolean;
    tw, th: DWORD;
    {$IFDEF test_SetField}
    test: DWORD;
    {$ENDIF}
begin
  Result := FALSE;
  Clear;
  //##################### _TIFFOpenEx
  tif := AllocMem( SizeOf( TTiffInfo ) + Length( fLoadingFromFile ) + 1 );
  tif.tif_name := PChar( DWORD( tif ) + Sizeof( TTiffInfo ) );
  strcopy( tif.tif_name, PChar( fLoadingFromFile ) );
  tif.tif_mode := 0 {O_RDONLY};
  tif.tif_curdir := WORD( -1 ); // not-existent directory
  tif.tif_curoff := 0;
  tif.tif_curstrip := DWORD( -1 ); // invalid strip
  tif.tif_row := DWORD( -1 ); // read pre-increment
  tif.tif_clientdata := Strm;
  tif.tif_readproc := @ readproc;
  tif.tif_writeproc := @ writedummy;
  tif.tif_seekproc := @ seekproc;
  tif.tif_closeproc := @ closeproc;
  tif.tif_sizeproc := @ sizeproc;
  tif.tif_mapproc := @ mapdummy;
  tif.tif_unmapproc := @ unmapdummy;
  j___TIFFSetDefaultCompressionState(Integer(tif));
  (*
   * Default is to return data MSB2LSB and enable the
   * use of memory-mapped files and strip chopping when
   * a file is opened read-only.
   *)
  tif.tif_flags := 1 {FILLORDER_MSB2LSB};
  //if (m == O_RDONLY )
  {if (Strm.Handle <> 0) and (Strm.Handle <> INVALID_HANDLE_VALUE) then
    tif.tif_flags := tif.tif_flags or TIFF_MAPPED;}
  tif.tif_flags := tif.tif_flags or TIFF_STRIPCHOP;
  if not ReadOK( tif.tif_header, sizeof( TTIFFHeader ) ) then Exit;
  (*
   * Setup the byte order handling.
   *)
  if ((tif.tif_header.tiff_magic <> TIFF_BIGENDIAN) and
      (tif.tif_header.tiff_magic <> TIFF_LITTLEENDIAN)) then
  begin
    //TIFFError(name,  "Not a TIFF file, bad magic number %d (0x%x)",
    //    tif->tif_header.tiff_magic,
    //    tif->tif_header.tiff_magic);
    Exit;
  end;
  InitOrder( tif.tif_header.tiff_magic );
  (*
   * Swap header if required.
   *)
  if (tif.tif_flags and TIFF_SWAB <> 0) then
  begin
    j__TIFFSwabShort(Integer( @ tif.tif_header.tiff_version ));
    j__TIFFSwabLong(Integer( @ tif.tif_header.tiff_diroff ));
  end;
  (*
   * Now check version (if needed, it's been byte-swapped).
   * Note that this isn't actually a version number, it's a
   * magic number that doesn't change (stupid).
   *)
  if (tif.tif_header.tiff_version <> TIFF_VERSION) then
  begin
    {TIFFError(name,
        "Not a TIFF file, bad version number %d (0x%x)",
        tif->tif_header.tiff_version,
        tif->tif_header.tiff_version);}
    Exit;
  end;
  tif.tif_flags := tif.tif_flags or TIFF_MYBUFFER;
  tif.tif_rawcp := nil;
  tif.tif_rawdata := 0;
  tif.tif_rawdatasize := 0;
  (*
   * Setup initial directory.
   *)
  tif.tif_nextdiroff := tif.tif_header.tiff_diroff;
  (*
   * Try to use a memory-mapped file if the client
   * has not explicitly suppressed usage with the
   * 'm' flag in the open mode (see above).
   *)
  {if ((tif->tif_flags & TIFF_MAPPED) &&
!TIFFMapFileContents(tif, (tdata_t*) &tif->tif_base, &tif->tif_size))
          tif->tif_flags &= ~TIFF_MAPPED;}
  if (j__TIFFReadDirectory(Integer( tif )) <> 0) then
  begin
    tif.tif_rawcc := DWORD( -1 );
    tif.tif_flags := tif.tif_flags or TIFF_BUFFERSETUP;
  end;
  //###################### CxImageTIFF::Decode
  bitspersample := 1;
  samplesperpixel := 1;
  rowsperstrip := -1;
  photometric := 0;
  compression := 1;
  orientation := ORIENTATION_TOPLEFT;
  TRY
    info.nNumFrames := j__TIFFNumberOfDirectories( Integer( tif ) );
    fNumFrames := info.nNumFrames;
    info.nFrame := fFrame2Decode;
    if (fFrame2Decode < 0) or (fFrame2Decode >= fNumFrames) or
       (j__TIFFSetDirectory( Integer(tif), info.nFrame ) = 0) then
    begin
      // page not present in TIFF file
      Exit;
    end;
    //get image info
    GetField( TIFFTAG_IMAGEWIDTH, @ fWidth);
    GetField( TIFFTAG_IMAGELENGTH, @ fHeight);
    GetField( TIFFTAG_SAMPLESPERPIXEL, @ samplesperpixel);
    GetField( TIFFTAG_BITSPERSAMPLE, @ bitspersample);
    GetField( TIFFTAG_ROWSPERSTRIP, @ rowsperstrip);
    GetField( TIFFTAG_PHOTOMETRIC, @ photometric);
    GetField( TIFFTAG_ORIENTATION, @ orientation);

    rwidth := fWidth;
    rheight := fHeight;
    if fReadSingleTile then
    begin
      if j__TIFFIsTiled(Integer(tif)) > 0 then
      begin
        GetField( TIFFTAG_TILEWIDTH, @ fWidth);
        GetField( TIFFTAG_TILELENGTH, @ fHeight);
      end
        else
      begin
        if rowsperstrip < 0 then
          fReadSingleTile := FALSE
        else
          fHeight := rowsperstrip;
      end;
      if fReadSingleTile then
      begin
        tif.tif_read_single_tile_only := 1;
        tif.tif_read_tile_xcol := fReadXCol;
        tif.tif_read_tile_yrow := fReadYRow;
      end;
    end
      else
    begin
      tif.tif_read_single_tile_only := 0;
    end;

    if (info.nEscape = -1) or fGetDimensionOnly then
    begin
      GetField( TIFFTAG_COMPRESSION, @compression);
      fIsJpeg := //(compression = COMPRESSION_OJPEG) or
                 (compression = COMPRESSION_JPEG);
            // Return output dimensions only
            {head.biWidth = width;
            head.biHeight = height;
            throw "output dimensions returned";}
      Result := TRUE;
      Exit;
    end;

    GetFieldDefaulted( TIFFTAG_RESOLUTIONUNIT, @ res_unit);
    if GetField( TIFFTAG_XRESOLUTION, @ resolution) then
    begin
      if (res_unit = RESUNIT_CENTIMETER) then
        resolution := (resolution*2.54 + 0.5);
      SetXDPI(Round( resolution ));
    end;
    if GetField( TIFFTAG_YRESOLUTION, @ resolution) then
    begin
      if (res_unit = RESUNIT_CENTIMETER) then
        resolution := (resolution*2.54 + 0.5);
      SetYDPI(Round( resolution ));
    end;

    if GetField( TIFFTAG_XPOSITION, @ offset) then
      info.xOffset := Round( offset );
    if GetField( TIFFTAG_YPOSITION, @ offset) then
      info.yOffset := Round( offset );

    head.biClrUsed := 0;
    info.nBkgndIndex := -1;

    if (rowsperstrip>Height) then
    begin
      rowsperstrip := height;
      j__TIFFSetField(Integer(tif), TIFFTAG_ROWSPERSTRIP, Pointer(rowsperstrip));
      {$IFDEF test_SetField}
      j__TIFFGetField(Integer(tif), TIFFTAG_ROWSPERSTRIP,@test);
      if test <> DWORD( rowsperstrip ) then
      asm
        nop
      end;
      {$ENDIF}
    end;

    isRGB := {(bitspersample >= 8) and} (photometric = PHOTOMETRIC_RGB) or
            (photometric = PHOTOMETRIC_YCBCR) or
            (photometric = PHOTOMETRIC_SEPARATED) or
            (photometric = PHOTOMETRIC_LOGL) or
            (photometric = PHOTOMETRIC_LOGLUV);

    if (isRGB) then
      head.biBitCount := 24
    else
    begin
      if ((photometric=PHOTOMETRIC_MINISBLACK)or
          (photometric=PHOTOMETRIC_MINISWHITE) or
          (photometric=PHOTOMETRIC_PALETTE)) then
      begin
        if (bitspersample = 1) then
        begin
          head.biBitCount :=1;		//B&W image
          head.biClrUsed  :=2;
        end else if (bitspersample = 4) then
        begin
          head.biBitCount :=4;		//16 colors gray scale
          head.biClrUsed  :=16;
        end else
        begin
          head.biBitCount :=8;		//gray scale
          head.biClrUsed  :=256;
        end;
      end
      else if (bitspersample = 4) then
      begin
        head.biBitCount :=4;			// 16 colors
        head.biClrUsed  :=16;
      end else
      begin
        head.biBitCount :=8;			//256 colors
        head.biClrUsed  :=256;
      end;

      if (bitspersample > 8) and (photometric=PHOTOMETRIC_PALETTE) then
      begin
        head.biBitCount :=24;
        head.biClrUsed  :=0;
      end;

    end;

    if (info.nEscape <> 0) then
    begin
      //throw "Cancelled"; // <vho> - cancel decoding
      Exit;
    end;

    if rowsperstrip < 0 then
      rowsperstrip := fHeight;

    GetField( TIFFTAG_COMPRESSION, @compression);
    if (fJpegScale <> jsFullSize) and
       ((compression = COMPRESSION_OJPEG) or
        (compression = COMPRESSION_JPEG)) then
    begin
      tif.tif_jpeg_scale := Integer( fJpegScale );
      fWidth := fWidth  div (1 shl tif.tif_jpeg_scale);
      fHeight:= fHeight div (1 shl tif.tif_jpeg_scale);
      rwidth := rwidth  div (1 shl tif.tif_jpeg_scale);
      rheight:= rheight div (1 shl tif.tif_jpeg_scale);
      rowsperstrip := rowsperstrip div (1 shl tif.tif_jpeg_scale);
      j__TIFFSetField( Integer(tif), TIFFTAG_ROWSPERSTRIP, Pointer( rowsperstrip ));
      {$IFDEF test_SetField}
      j__TIFFGetField(Integer(tif), TIFFTAG_ROWSPERSTRIP,@test);
      if test <> DWORD( rowsperstrip ) then
      asm
        nop
      end;
      {$ENDIF}

      if j__TIFFIsTiled( Integer( tif )) > 0 then
      begin
        j__TIFFGetField(Integer(tif), TIFFTAG_TILEWIDTH, @tw);
        j__TIFFGetField(Integer(tif), TIFFTAG_TILELENGTH, @th);
        tw := tw div (1 shl tif.tif_jpeg_scale);
        th := th div (1 shl tif.tif_jpeg_scale);
        j__TIFFSetField(Integer(tif), TIFFTAG_TILEWIDTH, Pointer(tw));
        {$IFDEF test_SetField}
        j__TIFFGetField(Integer(tif), TIFFTAG_TILEWIDTH,@test);
        if test <> tw then
        asm
          nop
        end;
        {$ENDIF}
        j__TIFFSetField(Integer(tif), TIFFTAG_TILELENGTH, Pointer(th));
        {$IFDEF test_SetField}
        j__TIFFGetField(Integer(tif), TIFFTAG_TILELENGTH,@test);
        if test <> th then
        asm
          nop
        end;
        {$ENDIF}
      end;
    end
    else
    begin
      fJpegScale := jsFullSize;
      tif.tif_jpeg_scale := 0;
    end;

    CreateImage(fWidth,fHeight,head.biBitCount, 0 {CXIMAGE_FORMAT_TIF} );	//image creation
    if (pDib = nil) then
    begin
      //throw "CxImageTIF can't create image";
      Exit;
    end;

    if samplesperpixel = 4 then
      AlphaCreate; //add alpha support for 32bpp tiffs
    if (samplesperpixel = 2) and (bitspersample = 8) then
      AlphaCreate; //add alpha support for 8bpp + alpha

    SetCodecOption(compression, 0); // <DPR> save original compression type

    if (isRGB) then
    begin
      Result := DecodeRGB;
    end
    else // not isRGB
    begin
      Result := DecodeWithPalette;
    end;
  EXCEPT
    Result := FALSE;
  END;
  j__TIFFClose( Integer(tif) );

  tif := nil;

end;

procedure TTIFF.MakeThumbnail(Bmp: PBitmap);
var R: TRect;
    y, y0, y1: Integer;
    Tmp, Tmp1, Tmp2, LastUsed: PBitmap;

    function ProvideBmp( h: Integer ): PBitmap;
    begin
      Result := Tmp1;
      TRY
        if (Result <> nil) and (Result.Height = h) then Exit;
        Result := Tmp2;
        if (Result <> nil) and (Result.Height = h) then Exit;
        if (LastUsed = Tmp1) or (Tmp1 = nil) then
        begin
          Free_And_Nil( Tmp1 );
          Tmp1 := NewDibBitmap( Width, h, pf16bit );
          Result := Tmp1;
        end;
        if Result = nil then
        begin
          Free_And_Nil( Tmp2 );
          Tmp2 := NewDibBitmap( Width, h, pf16bit );
          Result := Tmp2;
        end;
      FINALLY
        LastUsed := Result;
      END;
    end;
begin
  if empty then Exit;
  if (Bmp.Height >= Height) or (Bmp.Width >= Width) then
  begin
    R := Bmp.BoundsRect;
    StretchDibits( Bmp.Canvas.Handle, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
      0, 0, Width, Height, GetBits( 0 ), PBitmapInfo( pDib )^, DIB_RGB_COLORS,
      SRCCOPY );
    Exit;
  end;
  Tmp1 := nil;
  Tmp2 := nil;
  LastUsed := nil;
  TRY
    for y := 0 to Bmp.Height-1 do
    begin
      y0 := y * Height div Bmp.Height;
      y1 := (y+1) * Height div Bmp.Height;
      Tmp := ProvideBmp( y1 - y0 );
      Draw( Tmp.Canvas.Handle, 0, -y0 );
      Tmp.StretchDraw( Bmp.Canvas.Handle, MakeRect( 0, y, Bmp.Width, y+1 ) );
    end;
  FINALLY
    Tmp1.Free;
    Tmp2.Free;
  END;
end;

procedure TTIFF.Mirror;
begin
  //todo
end;

function TTIFF.PaletteSize: Integer;
begin
  Result := GetPaletteSize div Sizeof( TRGBQuad );
end;

function TTIFF.PixelFormat: TPixelFormat;
begin
  Result := pfDevice;
  if (pDib <> nil) then
  begin
    CASE (PBitmapInfoHeader( pDib ).biBitCount) OF
    1: Result := pf1bit;
    4: Result := pf4bit;
    8: Result := pf8bit;
    16: Result := pf16bit;
    24: Result := pf24bit;
    32: Result := pf32bit;
    END;
  end;
end;

function TTIFF.ReadOK(var buf; size: DWORD): Boolean;
begin
  Result := tif.tif_readproc( tif.tif_clientdata, @ buf, size ) = size;
end;


{
  (*
   * Setup header and write.
   *)
  if bigendian then
  begin
    if tif.tif_flags and TIFF_SWAB <> 0 then
      tif.tif_header.tiff_magic := TIFF_LITTLEENDIAN
    else
      tif.tif_header.tiff_magic := TIFF_BIGENDIAN;
  end
    else
  begin
    if tif.tif_flags and TIFF_SWAB <> 0 then
      tif.tif_header.tiff_magic := TIFF_BIGENDIAN
    else
      tif.tif_header.tiff_magic := TIFF_LITTLEENDIAN;
  end;
  tif.tif_header.tiff_version := TIFF_VERSION;
  if (tif.tif_flags and TIFF_SWAB <> 0) then
    _TIFFSwabShort( Integer( @ tif.tif_header.tiff_version ) );
  tif.tif_header.tiff_diroff := 0;	//* filled in later */
  (*
   * Setup the byte order handling.
   *)
  InitOrder(tif.tif_header.tiff_magic);
}

procedure TTIFF.RotateLeft;
begin
  //todo
end;

procedure TTIFF.RotateRight;
begin
  //todo
end;

function TTIFF.SetCodecOption(opt, imgtype: Integer): Boolean;
begin
  Result := TRUE;
  info.dwCodecOpt[imgtype] := opt;
  (*
  if (imagetype<CMAX_IMAGE_FORMATS){
          if (imagetype==0){
                  imagetype = GetType();
          }
          info.dwCodecOpt[imagetype] = opt;
          return true;
  }
  *)
end;

procedure TTIFF.SetPalette(pal: PRGBQuadArray; nColors: Integer);
begin
  if (pal = nil) or (pDib = nil) or (head.biClrUsed = 0 ) then Exit;
  move(pal^,GetPalette()^,min(GetPaletteSize(),nColors*sizeof(RGBQUAD)));
  info.last_c_isvalid := false;
end;

procedure TTIFF.SetPixelColor(x, y: Integer; c: TRGBQUAD;
  bSetAlpha: Boolean);
var iDst: PByte;
begin
  if (pDib = nil) or (x<0) or (y<0) or
     (x >= head.biWidth) or (y >= head.biHeight) then Exit;
  if (head.biClrUsed <> 0) then
    SetPixelIndex(x,y,GetNearestIndex(c))
  else
  begin
    iDst := @ info.pImage[ y*info.dwEffWidth + x*3 ];
    iDst^ := c.rgbBlue; inc( iDst );
    iDst^ := c.rgbGreen;inc( iDst );
    iDst^ := c.rgbRed;
    if (bSetAlpha) then
      //AlphaSet(x,y,c.rgbReserved);
      pAlpha[ x + y * Width ] := c.rgbReserved;
  end;
end;

procedure TTIFF.SetPixelIndex(x, y: Integer; i: Byte);
var pos: Byte;
    iDst: PByte;
begin
  if (pDib = nil) or (head.biClrUsed = 0) or
     (x<0) or (y<0) or
     (x >= head.biWidth) or (y >= head.biHeight) then Exit;

  if (head.biBitCount = 8) then
    info.pImage[y*info.dwEffWidth + x] := i
  else
  begin
    iDst := @ info.pImage[ y*info.dwEffWidth + (x*head.biBitCount shr 3) ];
    if (head.biBitCount = 4) then
    begin
      pos := 4*(1 - x mod 2);
      iDst^ := iDst^ and not ($F shl pos);
      iDst^ := iDst^ or ((i and $F) shl pos);
    end
      else
    if head.biBitCount = 1 then
    begin
      pos := 7 - x mod 8;
      iDst^ := iDst^ and not (1 shl pos);
      iDst^ := iDst^ or ((i and 1) shl pos);
    end;
  end;
end;

procedure TTIFF.SetXDPI(dpi: Integer);
begin
  if (dpi<=0) then dpi:=96;
  info.xDPI := dpi;
  head.biXPelsPerMeter := floor(dpi * 10000.0 / 254.0 + 0.5);
end;

procedure TTIFF.SetYDPI(dpi: Integer);
begin
  if (dpi<=0) then dpi:=96;
  info.yDPI := dpi;
  head.biYPelsPerMeter := floor(dpi * 10000.0 / 254.0 + 0.5);
end;

procedure TTIFF.StretchDraw(DC: HDC; const R: TRect);
begin
  if Empty then Exit;
  if StretchDIBits( DC, R.Left, R.Top, R.Right-R.Left, R.Bottom-R.Top, 0, 0,
    Width, Height, GetBits( 0 ), PBitmapInfo( pDib )^, DIB_RGB_COLORS, SRCCOPY )
    <> Height then
  begin
    asm
      nop
    end;
  end;
end;

procedure TTIFF.TileToStrip(_out, _in: PByte; rows, cols: Integer; outskew,
  inskew: Integer);
begin
  while rows > 0 do
  begin
    move( _in^, _out^, cols );
    inc( _out, outskew + cols );
    inc( _in, inskew + cols );
    dec( rows );
  end;
end;

function TTIFF.Transparent: Boolean;
begin
  Result := pAlpha <> nil;
end;

end.
